/*
 * Mach Operating System
 * Copyright (c) 1991, 1992 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	pmap.c,v $
 * Revision 2.5  92/05/21  17:20:30  jfriedl
 * 	Some cleanup for gcc warnings.
 * 	Appended 'U' to constants that would otherwise be signed.
 * 		Added prototypes.
 * 	[92/05/16            jfriedl]
 * 
 * Revision 2.4  92/02/19  13:56:44  elf
 * 	Moved all the M88K_* defines in here. 
 * 	Updated includes.
 * 	Merged up.
 * 	[92/01/22  18:53:13  danner]
 * 	XPR's for shootdown debugging.
 * 	[91/09/19  13:50:15  danner]
 * 
 * Revision 2.3  91/08/24  12:14:55  af
 * 	Taken from danner.
 * 
 * Revision 2.2.3.1  91/08/19  13:47:24  danner
 * 	replaced use of ffs with fls, since we don't care in which order
 * 	 we get the bits.
 * 	[91/08/06            danner]
 * 	Changed the num_pte_allocation code to allocate at least 90 pte
 * 	 pages (360 Kb). This prevents pte_page thrashing by 4 way
 * 	 parallel piped compiles (4 * 4 = 16 active processes =~ 48 ptes)
 * 	 and nettrek servers ( 2 + 20 = 22 active processes =~ 66 ptes). 
 * 	[91/08/06            danner]
 * 	Changes from jfriedl, including 2.5 compatability
 * 	[91/08/06  11:49:37  danner]
 * 
 * 	Volatilized pmap_update_interrupt.
 * 	[91/07/19            danner]
 * 
 * 	Volatilized shootdown_wait().
 * 	[91/07/19            danner]
 * 
 * 	Fixed a bug in the shootdown code. The following is a mail
 * 	 message I sent about the bug.
 * 
 * 	From: Daniel.Stodolsky@cs.cmu.edu 
 * 	To: mach3@cs.cmu.edu 
 * 	Subject: TLB shootdown problem
 * 
 * 	Folks -
 * 
 * 		I'd like to bring to general awareness a problem (with
 * 	help from Rich Draves and Bob Baron) I've discovered the the
 * 	shootdown code as implemented in the Vax, Sequent and Luna88k
 * 	pmap modules. All of these modules use essentially the same code
 * 	based reported in the paper
 * 
 * 	"Translation Lookaside Buffer Consistency: A Software Approach"
 * 	(available as CMU-CS-88-201).
 * 
 * 	In this paper, an algorithm is reported for TLB shootdown in
 * 	multiprocessor systems that lack hardware support for TLB
 * 	consistency or flushing of remote TLBS. In order to reduce the
 * 	frequency and expense of the shootdown operation, two
 * 	optimizations are made.
 * 
 * 	(1). Shootdown requests are only sent to processors that are not
 * 	idle (i.e, not currently running the idle thread).
 * 
 * 	(2) An idle processor that receives a shootdown interrupt ignores
 * 	it until.
 * 
 * 	(3) When a processor becomes active (nonidle), it checks to see
 * 	if there are any outstanding TLB flush requests.
 * 
 * 	The check and state transitions are typically implemented by the
 * 	MARK_CPU_IDLE and MARK_CPU_ACTIVE macros (see kern/sched_prim.c,
 * 	vax/pmap.h).
 * 
 * 	These optimizations are valid and have shown themselves to be
 * 	substantial performance wins while maintaining the consistency of
 * 	user pmaps.
 * 
 * 	However, the IDLE/ACTIVE technology misses an important case -
 * 	interrupts. A device interrupt could be serviced from a processor
 * 	marked idle, which could use an invalidated mapping of the kernel
 * 	pmap.  (In general, it is safe to assume that the interrupt
 * 	handlers to not access user virtual memory).
 * 
 * 	An example is given below for a three processor system running
 * 	3.0:
 * 
 * 	(1) processor one is processes a disk device_read request from a
 * 	user task. A buffer is allocated at kernel virtual address K and
 * 	bound to physical address P1. pmap_enter is called to enter the
 * 	mapping Processor one queues a request to the disk, and blocks
 * 	the thread.  All processors (zero, one, two) are idle.
 * 
 * 	(2) Processor zero, while running the idle thread, receives an
 * 	interrupt from the disk. It reads the data into the buffer at
 * 	address K (this driver uses virtual, not physical addresses),
 * 	causing the mapping (K,P) to be placed in its TLB. It finishes
 * 	the io request and calls iodone.  This causes the io_done thread
 * 	to become runable. Processor zero finishes the interrupt and
 * 	returns to the idle loop.
 * 
 * 	(3) While processor zero is returning to the idle_loop, processor
 * 	one notices the io_done is runable, and starts running it. This
 * 	leads to the blocked device_read thread to be awoken. Processor
 * 	two picks up this thread, and processor one returns to the idle
 * 	loop.
 * 
 * 	(4) The awoken device_read thread (on processor two) copies the
 * 	memory out to user space. This results in a pmap_remove call on
 * 	the kernel virtual address K. The pmap module notices that
 * 	processors zero and one are idle, so it queues the invalidation
 * 	requests for them. Processor two invalidates its own TLB and
 * 	completes the return of the device_read request.
 * 
 * 	(4). A second disk device_read request comes in by the same
 * 	thread that made the first request (still running on processor
 * 	two). Again, buffer space is allocated for read request. Kernel
 * 	virtual address K is reused but bound to physical address P2. The
 * 	disk i/o request is queued, and the t
 * 
 * Revision 2.2  91/07/09  23:19:06  danner
 * 	Added pmap_virtual_space, pmap_free_pages, and pmap_next_page.
 * 	[91/06/25            danner]
 * 
 * 	Modified to protect the page which has the exception vector 
 * 	     table. (SYSV_BASE).
 * 	[91/05/22            danner]
 * 
 * 	ASSERT -> assert
 * 	[91/05/10            danner]
 * 
 * 	Removed mach_xp include and conditionals.
 * 	[91/05/07            danner]
 * 
 * Revision 2.4  91/04/10  16:06:10  mbj
 * 	Added shootdown_wait, and updated PMAP_UPDATE_TLBS accordingly.
 * 	[91/04/09            danner]
 * 
 * 	Changed uses of cmmu_set to cmmu_set_batc_entry.
 * 	[91/03/29            danner]
 * 
 * 	Added code to count shootdown events by which cpu is shooting
 * 	 down.
 * 	[91/03/20            danner]
 * 
 * Revision 2.3  91/04/05  13:56:10  mbj
 * 	Initial code from the Omron 1.10 kernel release corresponding to X130.
 * 	The Copyright has been adjusted to correspond to the understanding
 * 	between CMU and the Omron Corporation.
 * 	[91/04/04            rvb]
 * 
 * 	Added pmap_extract_unlocked for the debugger.
 * 	[91/03/13            danner]
 * 
 * 	Changed mach/luna to  mach/luna88k
 * 	[91/03/08            danner]
 * 
 */

/*
 *	Manages physical address maps.
 *
 *	In addition to hardware address maps, this
 *	module is called upon to provide software-use-only
 *	maps which may or may not be stored in the same
 *	form as hardware maps.  These pseudo-maps are
 *	used to store intermediate results from copy
 *	operations to and from address spaces.
 *
 *	Since the information managed by this module is
 *	also stored by the logical address mapping module,
 *	this module may throw away valid virtual-to-physical
 *	mappings at almost any time.  However, invalidations
 *	of virtual-to-physical mappings must be done as
 *	requested.
 *
 *	In order to cope with hardware architectures which
 *	make virtual-to-physical map invalidates expensive,
 *	this module may delay invalidate or reduced protection
 *	operations until such time as they are actually
 *	necessary.  This module is given full information as
 *	to which processors are currently using which maps,
 *	and to when physical maps must be made correct.
 */

#include <mach/machine/vm_types.h>

#include <mach/boolean.h>
#include <kern/thread.h>
#include <kern/zalloc.h>

#include <kern/lock.h>

#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <vm/vm_kern.h>
#include <mach/vm_param.h>
#include <mach/vm_prot.h>
#include <vm/vm_page.h>
#include <vm/vm_user.h>

#include <luna88k/machparam.h>
#include <motorola/m88k/m88100/m88100.h>
#include <motorola/m88k/m88100/m882xx.h>
#include <luna88k/board.h>
#include <mach/luna88k/vm_param.h>
#include <luna88k/pcb.h>
#include <luna88k/softint.h> /* proto */


/*
 *	MC88200 protection codes in description.
 */

#define M88K_PTE_PFN	0xfffff000U	/* page frame address */
#define M88K_PTE_WT	0x00000200U	/* write through */
#define M88K_PTE_SP	0x00000100U	/* supervisor protection */
#define M88K_PTE_GL	0x00000080U	/* global */
#define M88K_PTE_CI	0x00000040U	/* cache inhibit */
#define M88K_PTE_MOD	0x00000010U	/* modified */
#define M88K_PTE_USED	0x00000008U	/* used */
#define M88K_PTE_WP	0x00000004U	/* write protect */
#define M88K_PTE_VALID	0x00000001U	/* valid */

#define	M88K_PTE_UR	M88K_PTE_WP	
#define	M88K_PTE_UW	0
#define M88K_PTE_KR	(M88K_PTE_SP|M88K_PTE_WP)
#define M88K_PTE_KW	M88K_PTE_SP

#define M88K_PTE_WIRED	0x00000002U

#define M88K_STE_PTB	0xfffff000U	/* page table base address */
#define M88K_STE_WT	M88K_PTE_WT
#define M88K_STE_SP	M88K_PTE_SP
#define M88K_STE_GL	M88K_PTE_GL
#define M88K_STE_CI	M88K_PTE_CI
#define M88K_STE_WP	M88K_PTE_WP
#define M88K_STE_VALID	M88K_PTE_VALID

#define	M88K_STE_UR	M88K_PTE_UR	
#define	M88K_STE_UW	M88K_PTE_UW
#define M88K_STE_KR	M88K_PTE_KR
#define M88K_STE_KW	M88K_PTE_KW

#define M88K_AD_STB	0xfffff000U	/* segment table base address */
#define M88K_AD_WT	M88K_PTE_WT
#define M88K_AD_GL	M88K_PTE_GL
#define M88K_AD_CI	M88K_PTE_CI
#define M88K_AD_TE	M88K_PTE_VALID

#define M88K_PROT_BITS	(M88K_PTE_SP|M88K_PTE_WP)
#define M88K_PROT_SET(pte,prot)	(*(int *)(pte)=(*(int *)(pte)&~M88K_PROT_BITS)|(prot))
#define P0PAGES		0xC0000		/* number of pages in P0 (for 3G) */
#define P1PAGES		(0x40000 - 0x400)/* number of pages in P1 (for 1G - 4M) */


struct segment_map {
    segment_d_t sd_entry[1024];
};

#define	TEXT_WRITE_OK		/* Kernel text write OK for debugging */
/*
 *	Private data structures.
 */
/*
 *	For each vm_page_t, there is a list of all currently
 *	valid virtual mappings of that page.  An entry is
 *	a pv_entry_t; the list is the pv_table.
 */
typedef struct pv_entry {
	struct pv_entry	*next;		/* next pv_entry */
	pmap_t		pmap;		/* pmap where mapping lies */
	vm_offset_t	va;		/* virtual address for mapping */
} *pv_entry_t;

#define PV_ENTRY_NULL	((pv_entry_t) 0)

extern vm_offset_t	avail_start, avail_next, avail_end;
extern vm_offset_t	virtual_avail, virtual_end;

pv_entry_t	pv_head_table;		/* array of entries, one per page */
zone_t		pv_list_zone;		/* zone of pv_entry structures */
int		pv_list_base;		/* pre-alloc pv_list 1/17/91 XXX */

/*
 *	Each entry in the pv_head_table is locked by a bit in the
 *	pv_lock_table.  The lock bits are accessed by the physical
 *	address of the page they lock.
 */
typedef volatile char *byte_lock_t;

byte_lock_t	pv_lock_table;
#define	pv_lock_table_size(n)	((((n) + (sizeof (int)) - 1) / (sizeof (int))) * (sizeof (int)))
/*
 *	First and last physical addresses that we maintain any information
 *	for.  Initialized to zero so that pmap operations done before
 *	pmap_init won't touch any non-existent structures.
 */
vm_offset_t	vm_first_phys = (vm_offset_t) 0;
vm_offset_t	vm_last_phys  = (vm_offset_t) 0;

/*
 *	Index into pv_head table, its lock bits, and the modify bits
 *	starting at vm_first_phys.
 */

#define pa_index(pa)	(atop(pa - vm_first_phys))

#define pai_to_pvh(pai)		(&pv_head_table[pai])
#define	lock_pvh_pai(pai)	(byte_lock(&pv_lock_table[pai]))
#define	unlock_pvh_pai(pai)	(byte_unlock(&pv_lock_table[pai]))

/*
 *	Array of modify bits, one byte per physical page (to avoid expense
 *	of locking)
 */
char	*pmap_modify_list;

/*
 *	Physical memory is reserved for user page tables, and allocated
 *	one M88K page (4K bytes, 1024 ptes) at a time when necessary.
 */
struct pte_page {
	queue_chain_t	link;		/* link in active or free queues */
	pmap_t		pmap;		/* pmap this page is allocated to */
	vm_offset_t	va;		/* first virtual address mapped
					   by this page */
	int		use_count;	/* number of mappings in use */
	int		wired_count;	/* number of mappings wired in */
};
typedef	struct pte_page	*ptep_t;	/* pointer to one */
#define PTE_PAGE_NULL	((ptep_t)0)

ptep_t	ptep_array;			/* pointer to array of ptep_t
					   for each pte page allocated */

queue_head_t	pte_page_queue_active;	/* list of all active pte pages */
queue_head_t	pte_page_queue_free;	/* list of all free pte pages */

decl_simple_lock_data(,
		pte_page_queue_lock)	/* both queues are locked either
					   by this lock, or by a write-
					   lock on the pmap system lock. */

vm_offset_t	first_pte_pa;		/* physical address of first pte
					   page */
int		first_pte_pfn;		/* page number for above */
int		num_pte_pages = 0;	/* number of pte pages allocated */

#define pfn_to_pte_page(pfn)	(&ptep_array[(pfn) - first_pte_pfn])
#define pte_page_to_pfn(ptep)	(first_pte_pfn + ((ptep) - ptep_array))

/*
 *	One pte page holds 1024 ptes, each one of which maps 4K bytes.
 *	How many bytes can be mapped using ptes in one page? (= 4M bytes).
 */
#define pte_page_map_size	(m88k_ptob(M88K_PGBYTES/sizeof(pt_entry_t)))

#define round_pte_page(v)	(((v)+M88K_PGBYTES-1)&~(M88K_PGBYTES-1))

/*
 *	Statistics
 */
int	pte_page_active_count = 0;
int	pte_page_free_count   = 0;
int	pte_page_wired_count  = 0;
int	pte_page_max_active   = 0;
int	pte_page_max_wired    = 0;

/*
 *	Locking and TLB invalidation
 */

/*
 *	Locking Protocols:
 *
 *	There are two structures in the pmap module that need locking:
 *	the pmaps themselves, and the per-page pv_lists (which are locked
 *	by locking the pv_lock_table entry that corresponds to the pv_head
 *	for the list in question.)  Most routines want to lock a pmap and
 *	then do operations in it that require pv_list locking -- however
 *	pmap_remove_all and pmap_copy_on_write operate on a physical page
 *	basis and want to do the locking in the reverse order, i.e. lock
 *	a pv_list and then go through all the pmaps referenced by that list.
 *	To protect against deadlock between these two cases, the pmap_lock
 *	is used.  There are three different locking protocols as a result:
 *
 *  1.  pmap operations only (pmap_extract, pmap_access, ...)  Lock only
 *		the pmap.
 *
 *  2.  pmap-based operations (pmap_enter, pmap_remove, ...)  Get a read
 *		lock on the pmap_lock (shared read), then lock the pmap
 *		and finally the pv_lists as needed [i.e. pmap lock before
 *		pv_list lock.]
 *
 *  3.  pv_list-based operations (pmap_remove_all, pmap_copy_on_write, ...)
 *		Get a write lock on the pmap_lock (exclusive write); this
 *		also guaranteees exclusive access to the pv_lists.  Lock the
 *		pmaps as needed.
 *
 *	At no time may any routine hold more than one pmap lock or more than
 *	one pv_list lock.  Because interrupt level routines can allocate
 *	mbufs and cause pmap_enter's, the pmap_lock and the lock on the
 *	kernel_pmap can only be held at splvm.
 */

/*
 *	We raise the interrupt level to splvm, to block interprocessor
 *	interrupts during pmap operations.  We must take the CPU out of
 *	the cpus_active set while interrupts are blocked.
 */

#define	SPLVM(spl)	{ \
	spl = splvm(); \
	i_bit_clear(cpu_number(), &cpus_active, &cpus_active_lock); \
}
#define	SPLX(spl)	{ \
	i_bit_set(cpu_number(), &cpus_active, &cpus_active_lock); \
	splx(spl); \
}

/*
 *	Lock on pmap system
 */
lock_data_t	pmap_system_lock;

#define PMAP_READ_LOCK(pmap, spl) { \
	SPLVM(spl); \
	lock_read(&pmap_system_lock); \
	simple_lock(&(pmap)->lock); \
}

#define PMAP_WRITE_LOCK(spl) { \
	SPLVM(spl); \
	lock_write(&pmap_system_lock); \
}

#define PMAP_READ_UNLOCK(pmap, spl) { \
	simple_unlock(&(pmap)->lock); \
	lock_read_done(&pmap_system_lock); \
	SPLX(spl); \
}

#define PMAP_WRITE_UNLOCK(spl) { \
	lock_write_done(&pmap_system_lock); \
	SPLX(spl); \
}

#define PMAP_WRITE_TO_READ_LOCK(pmap) { \
	simple_lock(&(pmap)->lock); \
	lock_write_to_read(&pmap_system_lock); \
}

#define LOCK_PVH(index)		(lock_pvh_pai(index))

#define UNLOCK_PVH(index)	(unlock_pvh_pai(index))

#define PMAP_UPDATE_TLBS(pmap, s, e) pmap_update_tlbs(pmap,s,e)  

#define M88K_PATC_NUMBER	56

#define INVALIDATE_TLB(kernel, start, end){\
    register unsigned int _n;\
    _n = m88k_btopr((unsigned)(end) - (unsigned)(start));\
    if (_n < M88K_PATC_NUMBER){\
	register caddr_t _va; \
	for (_va = (start); _n > 0; _n--, _va += M88K_PGBYTES ){\
	    cmmu_flush_tlb((kernel), _va, M88K_PGBYTES);\
	}\
    } else {\
	cmmu_flush_tlb((kernel), 0, FLUSH_ALL);  /* Flush all */\
    }\
}


/*
 *	Structures to keep track of pending TLB invalidations
 */


#define UPDATE_LIST_SIZE 5    /* VAX: this value was 4 */

struct pmap_update_item {
    pmap_t		pmap;		/* pmap to invalidate */
    vm_offset_t	start;		/* start address to invalidate */
    vm_offset_t	end;		/* end address to invalidate */
} ;
typedef struct pmap_update_item *pmap_update_item_t;
/*
 *	List of pmap updates.  If the list overflows,
 *	the last entry is changed to invalidate all.
 */
struct pmap_update_list {
	simple_lock_data_t	lock;
	int			count;
	boolean_t		overflowed;
	struct pmap_update_item	item[UPDATE_LIST_SIZE];
};

typedef	struct pmap_update_list	*pmap_update_list_t;

struct pmap_update_list	cpu_update_list[NCPUS];

/*
 *	Allocator for virtual space for page tables.
 */
struct pt_region {
	struct pt_region	*next;
	struct pt_region	*prev;
	vm_offset_t		start;
	vm_offset_t		end;
	pmap_t			pmap;
};
typedef struct pt_region	*pt_region_t;
#define PT_REGION_NULL		((pt_region_t)0)

/*
 *	Use a pt_region for the head of the list to simplify
 *	boundary checking.
 */
struct pt_region	user_pt_list;
vm_offset_t		user_pt_min,
			user_pt_max;
vm_size_t		user_pt_total_size = 2*(P0PAGES*sizeof(pt_entry_t));
					/*
					 * Allow space to map just ONE full
					 * address space.
					 */
vm_size_t		user_pt_free_size;
pt_region_t		user_pt_last_in_use;

decl_simple_lock_data(,	user_pt_lock_data)
#if !defined(JEFF_DEBUG)
#  define user_pt_lock()		simple_lock(&user_pt_lock_data)
#  define user_pt_unlock()		simple_unlock(&user_pt_lock_data)
#  define USER_PT_LOCK_MARK

#else

# define USER_PT_LOCK_MARK user_pt_lock_line = __LINE__;
# include "luna88k/asm_macro.h"
  int user_pt_waiting_thread = -1;/*jeff*/
  int user_pt_waiting_line = -1;/*jeff*/
  int user_pt_lock_thread = -1;/*jeff*/
  int user_pt_lock_line = -1;/*jeff*/
  decl_simple_lock_data(,jeffs_lock);
# define user_pt_lock()	 {						\
    psr_type psr;							\
    int have_lock;							\
									\
    do {								\
      psr = disable_interrupts_return_psr();				\
      simple_lock(&jeffs_lock);						\
      user_pt_waiting_thread = (unsigned)current_thread();		\
      user_pt_waiting_line   = __LINE__;				\
      have_lock = simple_lock_try(&user_pt_lock_data);			\
      if (!have_lock) {							\
	simple_unlock(&jeffs_lock);					\
	set_psr(psr);							\
	while(*(volatile int *)&user_pt_lock_data) /*empty*/ ;		\
      }									\
    } while(!have_lock);						\
									\
    user_pt_waiting_thread = user_pt_waiting_line   = -1;		\
    user_pt_lock_thread = (unsigned)current_thread();			\
    USER_PT_LOCK_MARK;							\
									\
    simple_unlock(&jeffs_lock);						\
    set_psr(psr);							\
 }

# define user_pt_unlock() {					\
	psr_type psr = disable_interrupts_return_psr();		\
	simple_lock(&jeffs_lock);				\
								\
	user_pt_lock_thread = user_pt_lock_line = -1;		\
	simple_unlock(&user_pt_lock_data);			\
								\
	simple_unlock(&jeffs_lock);				\
	set_psr(psr);						\
 }
#endif

static void		user_pt_init();
static void		user_pt_destroy();
static void		user_pt_allocate();

#ifndef	DEBUG_USER_PT
#define DEBUG_USER_PT	0
#endif	DEBUG_USER_PT

/*
 *	Internal extension to pmap structure:
 */
struct a_pmap {
	struct pmap		pmap;
	struct pt_region	region_0;
	struct pt_region	region_1;
};
typedef struct a_pmap	*a_pmap_t;

/*
 *	Other useful macros.
 */

#define luna88k_p0(addr)	(((int)(addr) & 0xC0000000U) != 0xC0000000U)
				/* stack bit not on -> p0 (3G bytes) */
#define	current_pmap()		(vm_map_pmap(current_thread()->task->map))
#define	pmap_in_use(pmap, cpu)	(((pmap)->cpus_using & (1 << (cpu))) != 0)

#define	luna88k_pfn(addr)	(luna88k_p0(addr) ? (m88k_btop(addr)):((m88k_btop(addr) & 0x3ffff)))
				/*
				 * page frame number, stripping off high 2
				 * bits if in P1.
				 */
#define	pmap_syspte(addr)	(&((pt_entry_t *)luna88k_vsbr)[m88k_btop(addr)])
#define pmap_p0pte(pmap,addr)	(&(pmap)->p0br[luna88k_pfn(addr)])
#define pmap_p1pte(pmap,addr)	(&(pmap)->p1br[luna88k_pfn(addr)])


struct pmap	kernel_pmap_store;
pmap_t		kernel_pmap;
pt_entry_t	*luna88k_sbr;		/* physical address */
vm_size_t	luna88k_slr;
pt_entry_t	*luna88k_vsbr;		/* virtual address */
vm_offset_t	phys_map_vaddr1, phys_map_vaddr2;
pt_entry_t	*phys_map_pte1, *phys_map_pte2;

struct zone	*pmap_zone;		/* zone of pmap structures */

int		ptes_per_vm_page;	/* number of ptes required to map
					   one VM page. */

boolean_t	pmap_initialized = FALSE;/* Has pmap_init completed? */


static void pmap_remove_range();	/* forward */
static void signal_cpus();		/* forward */

struct user_segment_info {
    long	p0lr;		/* current user segment p0lr */
    long	p1lr;		/* current user segment p1lr */
};

struct user_segment_info user_segment_info[NCPUS];

#ifdef WANT_PROTOTYPES

 /* kern/debug.c */
 extern void panic(char*,...);

 /* kern/lock.c */
 extern void lock_done(lock_t);
 extern void lock_init(lock_t,boolean_t);
 extern void lock_read(lock_t);
 extern void lock_write(lock_t);
 extern void lock_write_to_read(lock_t);

 /* kern/printf.c */
 extern void printf(char*,...);

 /* kern/queue.c */
 extern int remque(struct queue_entry*);
 extern queue_entry_t dequeue_head(queue_t);
 extern void enqueue_head(queue_t,queue_entry_t);
 extern void enqueue_tail(queue_t,queue_entry_t);
 extern void insque(struct queue_entry*,struct queue_entry*);
 extern void remqueue(queue_t,queue_entry_t);

 /* kern/xpr.c */
 extern void xpr(char*,int,int,int,int,int);

 /* kern/zalloc.c */
 extern vm_offset_t zalloc(zone_t);
 extern void zcram(zone_t,vm_offset_t,vm_size_t);
 extern void zfree(zone_t,vm_offset_t);
 extern zone_t zinit(vm_size_t,vm_size_t,vm_size_t,boolean_t,char*);

 /* luna88k/locore/locore_asm_routines.c */
 extern boolean_t simple_lock_try(simple_lock_t);
 extern void blkclr(void*,unsigned int);
 extern void byte_lock(volatile unsigned char*);
 extern void byte_lock_init(volatile unsigned char*);
 extern void byte_unlock(volatile unsigned char*);
 extern void bzero(void*,unsigned int);
 extern void memcpy(void*,void*,unsigned int);
 extern void simple_lock(simple_lock_t);
 extern void simple_lock_init(simple_lock_t);
 extern void simple_unlock(simple_lock_t);

 /* luna88k/machdep.c */
 extern unsigned int splvm(void);
 extern unsigned int splx(unsigned int);

 /* luna88k/softint.c */
 extern void softcall(func_t,caddr_t,int);

 /* motorola/m88k/m88100/cmmu.c */
 extern void cmmu_flush_tlb(unsigned int,caddr_t,int);
 extern void cmmu_set_batc_entry(unsigned int,unsigned int,unsigned int);
 extern void cmmu_set_sapr(unsigned int);
 extern void cmmu_set_uapr(unsigned int);

 /* vm/vm_kern.c */
 extern kern_return_t kmem_alloc_pageable(vm_map_t,vm_offset_t*,vm_size_t);
 extern kern_return_t kmem_alloc_wired(vm_map_t,vm_offset_t*,vm_size_t);

 /* forwards */
 boolean_t pmap_is_modified(vm_offset_t);
 boolean_t pmap_is_referenced(vm_offset_t);
 boolean_t pmap_next_page(vm_offset_t*);
 boolean_t pmap_valid_page(vm_offset_t);
 boolean_t pmap_verify_free(vm_offset_t);
 pmap_t pmap_create(vm_size_t);
 pmap_t pmap_kernel(void);
 static boolean_t user_pt_find(pt_region_t,vm_size_t);
 static int shootdown_wait(pmap_t,int);
 static ptep_t pte_free_page_alloc(pmap_t,vm_offset_t);
 static ptep_t pte_page_alloc(pmap_t,vm_offset_t);
 static unsigned int fls(unsigned int);
 static vm_offset_t pmap_map(vm_offset_t,vm_offset_t,vm_offset_t,int);
 static vm_size_t pmap_size(vm_size_t);
 static void interrupt_processor(int);
 static void m88k_protection_init(void);
 static void pmap_copy_on_write(vm_offset_t);
 static void pmap_remove_all(vm_offset_t);
 static void pmap_remove_range(pmap_t,vm_offset_t,pt_entry_t*,pt_entry_t*,boolean_t);
 static void pte_page_free(ptep_t);
 static void pte_page_unwire(ptep_t);
 static void pte_page_wire(ptep_t);
 static void signal_cpus(cpu_set,pmap_t,vm_offset_t,vm_offset_t);
 static void user_pt_allocate(pmap_t,vm_offset_t);
 static void user_pt_delete(pt_region_t);
 static void user_pt_destroy(pt_region_t);
 static void user_pt_init(void);
 static void user_pt_move(pt_region_t,vm_offset_t,vm_size_t);
 static void user_pt_move_ptes(vm_offset_t,vm_offset_t,vm_size_t);
 unsigned int pmap_free_pages(void);
 vm_offset_t pmap_extract(pmap_t,vm_offset_t);
 vm_offset_t pmap_extract_unlocked(pmap_t,vm_offset_t);
 void pmap_activate(pmap_t,thread_t,int);
 void pmap_bootstrap(vm_offset_t,vm_offset_t*,vm_offset_t*,vm_offset_t*,vm_offset_t*);
 void pmap_change_wiring(pmap_t,vm_offset_t,boolean_t);
 void pmap_clear_modify(vm_offset_t);
 void pmap_clear_reference(vm_offset_t);
 void pmap_collect(pmap_t);
 void pmap_copy(pmap_t,pmap_t,vm_offset_t,vm_size_t,vm_offset_t);
 void pmap_copy_page(vm_offset_t,vm_offset_t);
 void pmap_destroy(pmap_t);
 void pmap_enter(pmap_t,vm_offset_t,vm_offset_t,vm_prot_t,boolean_t);
 void pmap_init(void);
 void pmap_page_protect(vm_offset_t,vm_prot_t);
 void pmap_pageable(pmap_t,vm_offset_t,vm_offset_t,boolean_t);
 void pmap_protect(pmap_t,vm_offset_t,vm_offset_t,vm_prot_t);
 void pmap_reference(pmap_t);
 void pmap_remove(pmap_t,vm_offset_t,vm_offset_t);
 void pmap_set_kernel_mmu(void);
 void pmap_set_user_segment(pmap_t);
 void pmap_update_interrupt(int);
 void pmap_update_tlbs(pmap_t,vm_offset_t,vm_offset_t);
 void pmap_virtual_space(vm_offset_t*,vm_offset_t*);
 void pmap_zero_page(vm_offset_t);
 void process_pmap_updates(pmap_t);
 void user_pt_collect(pt_region_t,vm_size_t);
#endif /* ifdef WANT_PROTOTYPES */

/*
 *	Assume that there are three protection codes, all using low bits.
 */
int	user_protection_codes[8];
int	kernel_protection_codes[8];
/*
 *	Cache policy
 */
int	cache_policy =   M88K_PTE_GL;
/*
 *	Convert machine independent protection code
 *	to MC88200 dependent protection code.
 */

static void m88k_protection_init()
{
	register int	*kp, *up, prot;


	kp = kernel_protection_codes;
	up = user_protection_codes;

	for (prot = 0; prot < 8; prot++) {
		switch (prot) {
		case VM_PROT_NONE | VM_PROT_NONE | VM_PROT_NONE:
			*kp++ = 0;
			*up++ = 0;
			break;
		case VM_PROT_READ | VM_PROT_NONE | VM_PROT_NONE:
		case VM_PROT_READ | VM_PROT_NONE | VM_PROT_EXECUTE:
		case VM_PROT_NONE | VM_PROT_NONE | VM_PROT_EXECUTE:
			*kp++ = M88K_PTE_KR | cache_policy;
			*up++ = M88K_PTE_UR | cache_policy;
			break;
		case VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_NONE:
		case VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_EXECUTE:
		case VM_PROT_READ | VM_PROT_WRITE | VM_PROT_NONE:
		case VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE:
			*kp++ = M88K_PTE_KW | cache_policy;
			*up++ = M88K_PTE_UW | cache_policy;
			break;
		}
	}
}

#define	m88k_protection(map, prot)	\
	(((map) == kernel_pmap) ? kernel_protection_codes[prot] : \
				 user_protection_codes[prot])
#ifdef unused
/*
 *	Given an offset and a map, compute the address of the
 *	pte.  If the address is invalid with respect to the map
 *	then PT_ENTRY_NULL is returned (and the map may need to grow).
 *
 *	This is only used internally.
 */
static pt_entry_t *pmap_pte(map, addr)
	register pmap_t		map;
	register vm_offset_t	addr;
{
	register long	pfn;

	pfn = luna88k_pfn(addr);
	/*
	 *	Check for p0 or p1 region.
	 */

	if (luna88k_p0(addr)) {
		if (pfn >= map->p0lr)
			return(PT_ENTRY_NULL);
		return(map->p0br + pfn);
	}
	else {
		if (pfn < map->p1lr)
			return(PT_ENTRY_NULL);
		return(map->p1br + pfn);
	}
}
#endif

/*
 *	Macro version for speed in this module (C version needed for external
 *	references.
 */
#define pmap_pte(map, addr) \
	(luna88k_p0(addr) ? \
		((luna88k_pfn(addr) >= map->p0lr) ? \
			PT_ENTRY_NULL : (map->p0br + luna88k_pfn(addr))) : \
		((luna88k_pfn(addr) < map->p1lr) ? \
			PT_ENTRY_NULL : (map->p1br + luna88k_pfn(addr))))

#ifndef DEBUG_PTE_PAGE
#define	DEBUG_PTE_PAGE	0
#endif	DEBUG_PTE_PAGE

#if	DEBUG_PTE_PAGE
void ptep_check(ptep)
	ptep_t	ptep;
{
	register pt_entry_t	*pte, *epte;
	int			ctu, ctw;

	/* check the use and wired counts */
	if (ptep == PTE_PAGE_NULL)
		return;
	pte = pmap_pte(ptep->pmap, ptep->va);
	epte = pte + M88K_PGBYTES/sizeof(pt_entry_t);
	ctu = 0;
	ctw = 0;
	while (pte < epte) {
		if (pte->pfn != 0) {
			ctu++;
			if (pte->wired)
				ctw++;
		}
		pte += ptes_per_vm_page;
	}

	if (ctu != ptep->use_count || ctw != ptep->wired_count) {
		printf("ptep %x use %d wired %d - actual use %d wired %d\n",
		    	ptep, ptep->use_count, ptep->wired_count, ctu, ctw);
		panic("pte count");
	}
}
#endif	DEBUG_PTE_PAGE
/*
 *	Map memory at initialization.  The physical addresses being
 *	mapped are not managed and are never unmapped.
 *
 *	For now, VM is already on, we only need to map the
 *	specified memory.
 */
static vm_offset_t pmap_map(virt, start, end, prot)
	register vm_offset_t	virt;
	register vm_offset_t	start;
	register vm_offset_t	end;
	register int		prot;
{
	register int		ps;

	ps = PAGE_SIZE;
	while (start < end) {
		pmap_enter(kernel_pmap, virt, start, prot, FALSE);
		virt += ps;
		start += ps;
	}
	return(virt);
}

/*
 *	min_virtual_size is the minimum amount of kernel address space
 *	that we want.  It is declared global so that it can easily be
 *	patched.
 */

long	min_virtual_size = 32*1024*1024;

static vm_size_t pmap_size(size)
	vm_size_t	size;
{
	register long	map_size;

	map_size = size/M88K_PGBYTES * 4;	/* nonsense */
	if (map_size*M88K_PGBYTES < min_virtual_size)
		map_size = min_virtual_size/M88K_PGBYTES;
	else
	if (map_size*M88K_PGBYTES > VM_MAX_KERNEL_ADDRESS)
	    panic("map_size is too large");

	return(map_size);
}

#define	NPTEPPG	(M88K_PGBYTES/sizeof(pt_entry_t))
#include <luna88k/board.h>
/*
 * LUNA Hardware physcal address map
 */
struct physmap {
    vm_offset_t		pm_phys;		/* Physcal address */
    int			pm_size;		/* Address space */
    int			pm_cache;		/* cache on/off */
};

struct physmap physmap[] = {
    PROM_ADDR,		PROM_SPACE,		1,	/* P-Rom */
    FUSE_ROM_ADDR,	FUSE_ROM_SPACE,		0,	/* Fuse-Rom */
    NVRAM_ADDR,		NVRAM_SPACE,		0,	/* VN-Ram */
    OBIO_PIO0_BASE,	M88K_PGBYTES,		0,	/* PIO-0 */
    OBIO_PIO1_BASE,	M88K_PGBYTES,		0,	/* PIO-1 */
    OBIO_SIO,		M88K_PGBYTES,		0,	/* SIO */
    OBIO_TAS,		M88K_PGBYTES,		0,	/* TAS */
    OBIO_CLOCK0,	M88K_PGBYTES,		0,	/* Clock */
    INT_ST_MASK0,	M88K_PGBYTES,		0,	/* IntrMaskreg */
    SOFT_INT0,		M88K_PGBYTES,		0,	/* SoftINTreg */
    SOFT_INT_FLAG0,	M88K_PGBYTES,		0,	/* SoftINTflag */
    RESET_CPU0,		M88K_PGBYTES,		0,	/* CPUResetreg */
    TRI_PORT_RAM,	TRI_PORT_RAM_SPACE,	0,	/* 3 port RAM */
    EXT_A_ADDR,		EXT_A_SPACE,		0,	/* Ext board-A */
    EXT_B_ADDR,		EXT_B_SPACE,		0,	/* Ext board-B */
    PC_BASE,		PC_SPACE,		0,	/* Ext PC board */
    MROM_ADDR,		MROM_SPACE,		0,	/* Mask ROM */
    BMAP_START,		BMAP_SPACE,		0,	/* Bitmap */
    BMAP_PALLET0,	M88K_PGBYTES,		0,	/* Pallet */
    BMAP_PALLET1,	M88K_PGBYTES,		0,	/* Pallet */
    BMAP_PALLET2,	M88K_PGBYTES,		0,	/* Pallet */
    BOARD_CHECK_REG,	M88K_PGBYTES,		0,	/* board check reg */
    BMAP_CRTC,		M88K_PGBYTES,		0,	/* CRTC */
    SCSI_ADDR,		M88K_PGBYTES,		0,	/* SCSI */
    LANCE_ADDR,		M88K_PGBYTES,		0,	/* LANCE */
    0,			0
};
/*
 * M88K CMMU Segment Map
 */
segment_d_t	*kernel_segment_map;		/* For Supervisor Area */
segment_d_t	*user_segment_maps;		/* For User Area */
pt_entry_t	*veqr_page_table;		/* For VEQR page table */
int		kernel_area_descriptor;		/* Kernel area descriptor */
#define	veqr_pte(va)	(&veqr_page_table[pmap_syspte(va)->pfn])
extern	int	rawmem_size;			/* physcal mem-size */

/*
 *	Bootstrap the system enough to run with virtual memory.
 *	Map the kernel's code and data, and allocate the system page table.
 *	Called with mapping OFF.  Page_size must already be set.
 *
 *	Parameters:
 *	load_start:	PA where kernel was loaded
 *	avail_start	PA of first available physical page
 *	avail_end	PA of last available physical page
 *	virtual_avail	VA of first available page (after kernel bss)
 *	virtual_end	VA of last available page (end of kernel address space)
 *
 *	&start_text	start of kernel text
 *	&etext		end of kernel text
 */
void pmap_bootstrap(load_start, boot_avail_start, boot_avail_end,
			boot_virtual_avail, boot_virtual_end)

	vm_offset_t	load_start;
	vm_offset_t	*boot_avail_start;	/* IN/OUT */
	vm_offset_t	*boot_avail_end;	/* IN/OUT */
	vm_offset_t	*boot_virtual_avail;	/* IN/OUT */
	vm_offset_t	*boot_virtual_end;	/* OUT */
{
	register pt_entry_t	*sbr;
	long			slr;
	vm_offset_t		vaddr;
	vm_offset_t		s_text, e_text;
	extern int		start_text, etext;
	int			n,m;
	struct physmap 		*physp;
	vm_offset_t		vs;
	vm_offset_t		ve;
	pt_entry_t		*ptee;
	segment_d_t		*ste,*este;
	register pt_entry_t	*pte;
	int			cache_flag;

	/*
	 *	Set ptes_per_vm_page for general use.
	 */
	ptes_per_vm_page = PAGE_SIZE / M88K_PGBYTES;

	/*
	 *	Initialize protection arrays.
	 */
	m88k_protection_init();

	/*
	 *	The kernel's pmap is statically allocated so we don't
	 *	have to use pmap_create, which is unlikely to work
	 *	correctly at this part of the boot sequence.
	 */

	kernel_pmap = &kernel_pmap_store;

	lock_init(&pmap_system_lock, FALSE);	/* NOT a sleep lock */

	simple_lock_init(&kernel_pmap->lock);
	/*
	 * Kernel segment map allocate
	 */
	(*boot_avail_start) = (vm_offset_t)round_page(*boot_avail_start);
	kernel_segment_map = (segment_d_t *)(*boot_avail_start);
	blkclr((char*)kernel_segment_map, sizeof(struct segment_map));
	(*boot_avail_start) += sizeof(struct segment_map);

	/*
	 * Make kernel area descriptor
	 */
	kernel_area_descriptor = (int)kernel_segment_map | cache_policy | AREA_D_TE;

	/*
	 * User segment map(s) allocate
	 */
	(*boot_avail_start) = (vm_offset_t)round_page(*boot_avail_start);
	user_segment_maps = (segment_d_t *)(*boot_avail_start);
	blkclr((char*)user_segment_maps, sizeof(struct segment_map) * NCPUS);
	(*boot_avail_start) += sizeof(struct segment_map) * NCPUS;

	/*
	 * VEQR segment page table allocate
	 */
	(*boot_avail_start) = (vm_offset_t)round_page(*boot_avail_start);
	veqr_page_table = (pt_entry_t *)(*boot_avail_start);
	blkclr((char*)veqr_page_table,m88k_btop(rawmem_size)*sizeof(pt_entry_t));
	(*boot_avail_start) += m88k_btop(rawmem_size) * sizeof(pt_entry_t);
	(*boot_avail_start) = (vm_offset_t)round_page(*boot_avail_start);
	pv_list_base = (int)(*boot_avail_start);		/* XXX */
	(*boot_avail_start) += M88K_PGBYTES*NCPUS;		/* XXX */
	(*boot_avail_start) = (vm_offset_t)round_page(*boot_avail_start);

	/*
	 * VEQR Kernel segment and page table set
	 */
	ste = &kernel_segment_map[m88k_btos(VEQR_ADDR)];
	pte = veqr_page_table;
	for (n = 0; n < rawmem_size; n += (M88K_SGPAGES*M88K_PGBYTES)) {
	    *(int *)ste++ = M88K_STE_VALID | M88K_STE_KW |
		cache_policy | m88k_round_page(pte);
	    for (m = n; 
		 m < (n + M88K_SGPAGES*M88K_PGBYTES);
		 m+=M88K_PGBYTES) {
		*(int *)pte++ = M88K_PTE_VALID|M88K_PTE_KW|
		    cache_policy | m88k_round_page(m);
	    }
	}

	/*
	 * Physical IO address map page table allocate and set
	 */
	for (physp = physmap; physp->pm_phys != 0; physp++)
	{
	    /*
	     * Set data into segment table
	     */
	    vs = physp->pm_phys;
	    ve = vs + (physp->pm_size > M88K_SGBYTES ? physp->pm_size : M88K_SGBYTES);
	    if (!physp->pm_cache) {
		cache_flag = M88K_PTE_CI;
	    } else {
		cache_flag = (cache_policy & ~M88K_PTE_GL);
	    }

	    ste = &kernel_segment_map[m88k_btos(vs)];
	    este = &kernel_segment_map[m88k_btosr(ve)];
	
	    while (ste < este)
	    {

		/*
		 * Set segment table 
		 *	Segment table and Page table exist on
		 *	first area of physcal address equal vairtual.
		 *	Then we get page table address from segment table.
		 *	but segment entry has various flags.
		 */
		if (!((int)(pte = *(pt_entry_t **)ste) & M88K_STE_VALID)) {
		    /*
		     * Get Page Table
		     */
		    pte = (pt_entry_t *)(*boot_avail_start);
		    blkclr((char*)pte, M88K_PGBYTES);
		    *boot_avail_start += M88K_PGBYTES;
		    *(int *)ste = M88K_STE_VALID|M88K_STE_KW |
			cache_flag | m88k_round_page(pte);
		} else {
		    (int)pte &= M88K_PTE_PFN;
		}
		ptee = pte + M88K_PGBYTES / sizeof(pt_entry_t);
		
		/*
		 * Adjust pte
		 */
		pte += m88k_btop(vs - m88k_trunc_seg(vs));

		/*
		 * Set page table
		 */
		for (; vs < ve && pte < ptee; pte++, vs += M88K_PGBYTES) {
		    *(int *)pte = M88K_PTE_VALID | M88K_PTE_KW |
			cache_flag | m88k_round_page(vs);
		}
		ste++;
	    }
	}


	/*
	 *	Allocate the kernel page table from the front
	 *	of available physical memory.
	 */

	(*boot_avail_start) = (vm_offset_t)round_page(*boot_avail_start);
	sbr = kernel_pmap->p0br = (pt_entry_t *) (*boot_avail_start);
	slr = kernel_pmap->p0lr = pmap_size(*boot_avail_end - *boot_avail_start);
	kernel_pmap->ref_count = 1;
	luna88k_sbr = sbr;
	luna88k_slr = slr;

	for (n = 0; n < NCPUS; n++) {
	    user_segment_info[n].p0lr = 0;
	    user_segment_info[n].p1lr = P1PAGES;
	}

	/*
	 *	Set up p1br and p1lr to valid values, in case kernel_pmap
	 *	is loaded into a PCB.
	 */
	kernel_pmap->p1br = (pt_entry_t *)VM_MIN_KERNEL_ADDRESS;
	kernel_pmap->p1lr = P1PAGES;

	/*
	 *	Clear the entire table, thus making the length register OK.
	 */

	blkclr((char*)sbr, slr * sizeof(pt_entry_t));

	/*
	 *	Map low system memory, system text, system data/bss/symbols
	 */

	/*
	 *	&start_text and &etext can't have arithmetic done on them
	 *	directly because they must be physical addresses.
	 */

	s_text = (vm_offset_t) &start_text;
	e_text = (vm_offset_t) &etext;
	s_text = round_page(s_text);
	e_text = trunc_page(e_text);
	if (load_start != SYSV_BASE)
	    panic("load_start");
	vaddr = pmap_map(VM_MIN_ADDRESS, 
			 SYSV_BASE, s_text, VM_PROT_READ | VM_PROT_WRITE);

#ifdef	TEXT_WRITE_OK
	vaddr = pmap_map(vaddr, s_text, e_text, VM_PROT_READ|VM_PROT_WRITE);
#else  /* ! TEXT_WRITE_OK */
	vaddr = pmap_map(vaddr, s_text, e_text, VM_PROT_READ);
#endif /* ! TEXT_WRITE_OK */
	vaddr = pmap_map(vaddr, e_text, *boot_avail_start,
						VM_PROT_WRITE|VM_PROT_READ);

	/*
	 *	Map system page table
	 */

	*boot_avail_start = round_page(sbr + slr);
	*boot_virtual_avail = pmap_map(vaddr, (vm_offset_t) sbr, *boot_avail_start,
					VM_PROT_WRITE|VM_PROT_READ);
	*boot_virtual_end = VM_MIN_KERNEL_ADDRESS + (slr * M88K_PGBYTES);
	printf("Kernel virtual space from 0x00000000 to 0x%x.\n", *boot_virtual_end);

	/* 
	 * Make kernel segment table in kernel text/data/bss and
	 * heap area.
	 */
	ste = &kernel_segment_map[m88k_btos(VM_MIN_KERNEL_ADDRESS)];
	este = &kernel_segment_map[m88k_btos(*boot_virtual_end)];
	for (pte = sbr; ste <= este; ste++, pte += NPTEPPG) {
		*(int *)ste = M88K_STE_VALID | M88K_STE_KW |
		    cache_policy | m88k_round_page(pte);
	}
	/*
	 *	Once we turn on mapped mode, we need to reference the system
	 *	page table via its virtual address.  (Same for kernel_pmap).
	 */

	kernel_pmap->p0br = (pt_entry_t *) vaddr;
	luna88k_vsbr = (pt_entry_t *)vaddr;
	/*
	 * System page table must not be cached,
	 * becouse table walk do not use cache.
	 */
	pte = pmap_syspte(luna88k_sbr);
	ptee = veqr_pte(luna88k_sbr);
	for (n = 0; n < m88k_btopr(luna88k_slr*sizeof(pt_entry_t)); n++) {
	    pte->g = 1;
	    pte++ ->ci = 1;
	    ptee->g = 1;
	    ptee++ ->ci = 1;
	}
	pte = pmap_syspte(kernel_segment_map);
	ptee = veqr_pte(kernel_segment_map);
	for (n = 0; n < m88k_btopr(sizeof(struct segment_map)); n++) {
	    pte->g = 1;		
	    pte++ ->ci = 1;
	    ptee->g = 1;
	    ptee++ ->ci = 1;
	}
	pte = pmap_syspte(user_segment_maps);
	ptee = veqr_pte(user_segment_maps);
	for (n = 0; n < m88k_btopr(sizeof(struct segment_map) * NCPUS); n++) {
	    pte->g = 1;
	    pte++ ->ci = 1;
	    ptee->g = 1;
	    ptee++ ->ci = 1;
	}
	kernel_pmap = (pmap_t) (((vm_offset_t) kernel_pmap - load_start) +
				VM_MIN_KERNEL_ADDRESS);

	*boot_virtual_avail = round_page(*boot_virtual_avail);

	/*
	 *	Map the kernel into P0 space for the leap to mapped mode.
	 *	NB: this doesn't work for non-zero loadpoint!
	 *	    we're depending on the pre-fetch hardware.
	 */

	/*
	 *	Allocate some amount of physical memory to use for
	 *	user page tables.
	 */

	if (num_pte_pages == 0) {
		num_pte_pages = m88k_btop((*boot_avail_end - *boot_avail_start))/NPTEPPG;
		num_pte_pages *= 4;
		if (num_pte_pages < 90)
			num_pte_pages = 90;
	}

	first_pte_pa = *boot_avail_start;
	*boot_avail_start += round_page(m88k_ptob(num_pte_pages));
	blkclr((char *)first_pte_pa, m88k_ptob(num_pte_pages));
	first_pte_pfn = m88k_btop(first_pte_pa);
	ptee = &veqr_page_table[first_pte_pfn];
	for (n = 0; n < num_pte_pages; n++) {
	    ptee->g = 1;
	    ptee++ ->ci = 1;
	}

	/*
	 *	Don't forget to zero it.
	 */
	blkclr((caddr_t) first_pte_pa, *boot_avail_start - first_pte_pa);
	avail_next = *boot_avail_start;
}

/*
 *	Initialize the pmap module.
 *	Called by vm_init, to initialize any structures that the pmap
 *	system needs to map virtual memory.
 */
#ifdef MACH_KERNEL
void pmap_init()
#else
void pmap_init(phys_start, phys_end)
	vm_offset_t	phys_start, phys_end;
#endif
{
	register long		npages;
#ifdef MACH_KERNEL
	vm_offset_t	addr;
	vm_offset_t	phys_start = avail_start;
	vm_offset_t	phys_end   = avail_end;

#else
	register vm_offset_t	addr;
#endif

	register vm_size_t	s;
	byte_lock_t	pv_lock;
	int i;

	/*
	 *	Allocate memory for the pv_head_table and its lock bits,
	 *	the modify bit array, and the pte_page table.
	 */

	npages = atop(phys_end - phys_start);
	s = (vm_size_t) (sizeof(struct pv_entry) * npages
				+ pv_lock_table_size(npages)
				+ npages
				+ num_pte_pages * sizeof(struct pte_page));

	s = round_page(s);
#ifdef  MACH_KERNEL /* pure kernel */
	if (kmem_alloc_wired(kernel_map, &addr, s) != KERN_SUCCESS)
	  panic("pmap_init: kmem_alloc failure");
	blkclr((caddr_t) addr, s);
#else
	addr = (vm_offset_t) kmem_alloc(kernel_map, s);
	/*
	 *	memory was cleared by kmem-alloc
	 */
#endif

	/*
	 *	Allocate the structures first to preserve word-alignment.
	 */
	ptep_array = (ptep_t) addr;
	addr = (vm_offset_t)(ptep_array + num_pte_pages);

	pv_head_table = (pv_entry_t) addr;
	addr = (vm_offset_t) (pv_head_table + npages);

	pv_lock_table = (byte_lock_t) addr;
	/*
	 * Initialize the pv-list
	 */
	for (i = 0, pv_lock = pv_lock_table; i < npages; i++, pv_lock++) {
		byte_lock_init(pv_lock);
	}
	addr = (vm_offset_t) (pv_lock_table + pv_lock_table_size(npages));

	pmap_modify_list = (char *) addr;

	/*
	 *	Create the zone of physical maps,
	 *	and of the physical-to-virtual entries.
	 */
	s = (vm_size_t) sizeof(struct a_pmap);
	pmap_zone = zinit(s, 400*s, PAGE_SIZE, FALSE, "pmap");
	s = (vm_size_t) sizeof(struct pv_entry);
	pv_list_zone = zinit(s, 10000*s, PAGE_SIZE, FALSE, "pv_list");
	zcram(pv_list_zone, pv_list_base, M88K_PGBYTES*NCPUS);

	/*
	 *	Initialize the pte-page queues.
	 */
	simple_lock_init(&pte_page_queue_lock);
	queue_init(&pte_page_queue_free);
	queue_init(&pte_page_queue_active);

	{
	    register ptep_t	ptep;

	    for (i = 0, ptep = ptep_array; i < num_pte_pages; i++, ptep++) {
		enqueue_tail(&pte_page_queue_free, (queue_entry_t) ptep);
	    }

	}
	pte_page_free_count = num_pte_pages;

	/*
	 *	Set up the pmap request lists
	 */
	{
	    register pmap_update_list_t	up;

	    for (i = 0, up = cpu_update_list; i < NCPUS; i++, up++) {
		simple_lock_init(&up->lock);
		up->count = 0;
	up->overflowed = FALSE;
	    }
	}
	simple_lock_init(&cpus_active_lock);
	simple_lock_init(&cpus_idle_lock);

	/*
	 *	Initialize the user_pt allocation routines
	 */
	user_pt_init();

	/*
	 *	Only now, when all of the data structures are allocated,
	 *	can we set vm_first_phys and vm_last_phys.  If we set them
	 *	too soon, the kmem_alloc above will try to use these
	 *	data structures and blow up.
	 */

	vm_first_phys = phys_start;
	vm_last_phys = phys_end;

	pmap_initialized = TRUE;
}
/*
 *	The physical address space is dense... there are no holes.
 *	All addresses provided to vm_page_startup() are valid.
 */
boolean_t	pmap_valid_page(p)
	vm_offset_t	p;
{
	return(TRUE);
}

/*
 *	Does a page table entry actually exist?
 */
#define pte_page_valid(pte)	(pmap_syspte((vm_offset_t)(pte))->valid)

/*
 *	Given a pte, returns the pte page pointer for it.
 *	Does not work on kernel map.
 */
#define get_pte_page(pte) \
	(pfn_to_pte_page(pmap_syspte((vm_offset_t)(pte))->pfn))

/*
 *	Allocate a free pte page.
 *	The affected pmap must be locked.
 *	Returns NULL if there are no free pte pages.
 */
static ptep_t pte_free_page_alloc(pmap, va)
	register pmap_t		pmap;
	vm_offset_t		va;
{
	register ptep_t		ptep;
	vm_offset_t		sysva;
	pt_entry_t		*syspte;
	register int 		i;
	int			lock_flag = 0;

	simple_lock(&pte_page_queue_lock);
	if (queue_empty(&pte_page_queue_free)) {
		simple_unlock(&pte_page_queue_lock);
		return (PTE_PAGE_NULL);
	}

	ptep = (ptep_t) dequeue_head(&pte_page_queue_free);
	pte_page_free_count--;

	if (ptep->pmap != PMAP_NULL || ptep->va != (vm_offset_t) 0) {
		panic("ptep in use!");
	}

	/*
	 *	Enter the mapping for this page in the
	 *	system map.  The virtual address we enter is the
	 *	address that is mapped by the first pte in the
	 *	pte page.
	 */

	va &= ~(pte_page_map_size - 1);
	sysva = (vm_offset_t) pmap_pte(pmap, va);
	syspte = pmap_syspte(sysva);


	/*
	 *	Mapping is invalid - TLB invalidate is not needed
	 */
	if (pmap != kernel_pmap)
		simple_lock(&kernel_pmap->lock);
	else
		lock_flag = simple_lock_try(&kernel_pmap->lock);
	PMAP_UPDATE_TLBS(kernel_pmap, sysva, sysva + PAGE_SIZE/(sizeof(pt_entry_t)));

	*(int *)syspte = M88K_PTE_VALID | M88K_PTE_KW |
	    /* page table must be cache inhibit */
	    M88K_PTE_GL | M88K_PTE_CI |	
		m88k_ptob(pte_page_to_pfn(ptep));

	for (i = 0; i < NCPUS; i++)
	    bit_set(i, &(pmap->seg_changed));

	if (pmap != kernel_pmap || lock_flag)
		simple_unlock(&kernel_pmap->lock);

	lock_flag = simple_lock_try(&pmap->lock);
	PMAP_UPDATE_TLBS(pmap, (vm_offset_t)va, (vm_offset_t)va+pte_page_map_size);
	if (pmap == current_pmap() && pmap != kernel_pmap)
	    pmap_set_user_segment(pmap);
	if (lock_flag)
		simple_unlock(&pmap->lock);

	ptep->pmap = pmap;
	ptep->va = va;
	enqueue_tail(&pte_page_queue_active, (queue_entry_t) ptep);
	pte_page_active_count++;
	if (pte_page_active_count > pte_page_max_active)
		pte_page_max_active = pte_page_active_count;

	simple_unlock(&pte_page_queue_lock);

#if	DEBUG_PTE_PAGE
	ptep_check(ptep);	/* should be empty */
#endif	DEBUG_PTE_PAGE
	return(ptep);
}

/*
 *	Allocate a pte page.  May grab one from another pmap if
 *	there are no free pages.
 *	Must be called with the pmap system locked for write.
 */
static ptep_t pte_page_alloc(pmap, va)
	pmap_t		pmap;
	vm_offset_t	va;
{
	ptep_t	ptep;
	if (queue_empty(&pte_page_queue_free)) {

		pmap_t		old_pmap;
		vm_offset_t	old_va;
		pt_entry_t	*spte, *epte;

		/*
		 *	no free pte pages.  Recycle one.
		 */
		ptep = (ptep_t) queue_first(&pte_page_queue_active);
		if (ptep == PTE_PAGE_NULL) {
			panic("not enough pte pages!");
		}
		/*
		 *	free the pte page
		 */

		old_pmap = ptep->pmap;
		old_va = ptep->va;
		spte = pmap_pte(old_pmap, old_va);
		epte = spte + M88K_PGBYTES/sizeof(pt_entry_t);
		simple_lock(&old_pmap->lock);
		PMAP_UPDATE_TLBS(old_pmap, old_va, old_va + pte_page_map_size);
		pmap_remove_range(old_pmap, old_va, spte, epte, TRUE);
		/*
		 *	calls pte_page_free, which clears the pte page
		 *	and puts it at the head of the free queue.
		 */
		simple_unlock(&old_pmap->lock);
	}
	return (pte_free_page_alloc(pmap, va));
}
	
/*
 *	Free a pte page.
 *	The mappings in the pte page must have already been freed.
 *	The pmap system needs only to be locked for read.
 */
static void pte_page_free(ptep)
	register ptep_t	ptep;
{
	vm_offset_t	sysva;
	pt_entry_t	*syspte;
	register int     i;

#if	DEBUG_PTE_PAGE
	ptep_check(ptep);	/* should be empty */
#endif	DEBUG_PTE_PAGE
	/*
	 *	Remove the mapping from the system page table.
	 */
	sysva = (vm_offset_t) pmap_pte(ptep->pmap, ptep->va);
	syspte = pmap_syspte(sysva);
	simple_lock(&kernel_pmap->lock);
	PMAP_UPDATE_TLBS(kernel_pmap, sysva, sysva + M88K_PGBYTES);
	*(int *)syspte = 0;
	simple_unlock(&kernel_pmap->lock);

	/*
	 * Signal to other cpu for updating segment table entry.
	 */

	if (ptep->pmap) {
	    for (i = 0; i < NCPUS; i++)
		bit_set(i, &(ptep->pmap->seg_changed));
	    PMAP_UPDATE_TLBS(ptep->pmap, (vm_offset_t)0, (vm_offset_t)0);
	} else
	    printf("pte_page_free: ptep does not have pmap\n");

	/*
	 *	Clear the pte group fields.
	 */
	ptep->pmap = PMAP_NULL;
	ptep->va = (vm_offset_t) 0;
	ptep->use_count = 0;
	ptep->wired_count = 0;

	/*
	 *	Put it on the free list.
	 */
	simple_lock(&pte_page_queue_lock);
	remqueue(&pte_page_queue_active, (queue_entry_t) ptep);
	pte_page_active_count--;
	enqueue_head(&pte_page_queue_free, (queue_entry_t) ptep);
	pte_page_free_count++;
	simple_unlock(&pte_page_queue_lock);
}

/*
 *	Wire down a pte page so it cannot be reused.
 */
static void pte_page_wire(ptep)
	ptep_t	ptep;
{
	simple_lock(&pte_page_queue_lock);
	remqueue(&pte_page_queue_active, (queue_entry_t) ptep);
	pte_page_active_count--;
	pte_page_wired_count++;
	if (pte_page_wired_count > pte_page_max_wired)
		pte_page_max_wired = pte_page_wired_count;
	simple_unlock(&pte_page_queue_lock);
}

/*
 *	Unwire a pte page.
 */
static void pte_page_unwire(ptep)
	ptep_t	ptep;
{
	simple_lock(&pte_page_queue_lock);
	enqueue_tail(&pte_page_queue_active, (queue_entry_t) ptep);
	pte_page_wired_count--;
	pte_page_active_count++;
	simple_unlock(&pte_page_queue_lock);
}

/*
 *	Create and return a physical map.
 *
 *	If the size specified for the map
 *	is zero, the map is an actual physical
 *	map, and may be referenced by the
 *	hardware.
 *
 *	If the size specified is non-zero,
 *	the map will be used in software only, and
 *	is bounded by that size.
 */
pmap_t pmap_create(size)
	vm_size_t	size;
{
	register pmap_t		p;
	register pt_region_t	region;
	int	i;

	/*
	 *	A software use-only map doesn't even need a map.
	 */

	if (size != 0) {
		return(PMAP_NULL);
	}

	p = (pmap_t) zalloc(pmap_zone);
	/* ZALLOC(pmap_zone, p, pmap_t); */
	if (p == PMAP_NULL) {
		panic("pmap_create: cannot allocate a pmap");
	}

	/*
	 *	The VAX insists that p0br and p1br always point to
	 *	valid kernel-space addresses (or, in the case of p1br,
	 *	that the end of the "hole" in the p1 region must be
	 *	a valid kernel address).
	 */

	p->p0br = (pt_entry_t *)VM_MIN_KERNEL_ADDRESS;
	p->p0lr = 0;
	p->p1br = (pt_entry_t *)VM_MIN_KERNEL_ADDRESS;
	p->p1lr = P1PAGES;
	p->ref_count = 1;

	simple_lock_init(&p->lock);
	p->cpus_using = 0;
	p->seg_changed = 0;
	/* initialize block address translation cache */
	for (i = 0; i < NBATC; i++) {
	    p->i_batc[i].raw = 0;
	    p->d_batc[i].raw = 0;
	}

	/*
	 *	Initialize statistics.
	 */

	p->stats.resident_count = 0;
	p->stats.wired_count = 0;

	/*
	 *	Initialize the map regions to empty.
	 */
	region = &((a_pmap_t)p)->region_0;
	region->pmap = p;
	region->start = 0;
	region->end = 0;
	region->next = region;	/* chain region to itself */
	region->prev = region;	/* to mark it not-in-use */

	region = &((a_pmap_t)p)->region_1;
	region->pmap = p;
	region->start = 0;
	region->end = 0;
	region->next = region;	/* chain region to itself */
	region->prev = region;	/* to mark it not-in-use */

	return(p);
}

/*
 *	Retire the given physical map from service.
 *	Should only be called if the map contains
 *	no valid mappings.
 */

void pmap_destroy(p)
	register pmap_t	p;
{
	register int		c, s;

	if (p == PMAP_NULL)
		return;

	SPLVM(s);
	simple_lock(&p->lock);
	c = --p->ref_count;
	simple_unlock(&p->lock);
	SPLX(s);

	if (c != 0)
		return;	/* pmap still in use */

	user_pt_destroy(&((a_pmap_t)p)->region_0);
	user_pt_destroy(&((a_pmap_t)p)->region_1);

	zfree(pmap_zone, (vm_offset_t) p);
}

/*
 *	Add a reference to the specified pmap.
 */

void pmap_reference(p)
	register pmap_t	p;
{
	int	s;
	if (p != PMAP_NULL) {
		SPLVM(s);
		simple_lock(&p->lock);
		p->ref_count++;
		simple_unlock(&p->lock);
		SPLX(s);
	}
}

/*
 *	Remove a range of page-table entries.
 *	The entries given are the first (inclusive)
 *	and last (exclusive) entries for the VM pages.
 *	The virtual address is the va for the first pte.
 *
 *	The pmap must be locked.
 *	If the pmap is not the kernel pmap, the range must lie
 *	entirely within one pte-page.  This is NOT checked.
 *	Assumes that the pte-page exists.
 */

/* static */
static void pmap_remove_range(pmap, va, spte, epte, free_if_empty)
	pmap_t			pmap;
	vm_offset_t		va;
	pt_entry_t		*spte;
	pt_entry_t		*epte;
	boolean_t		free_if_empty;
{
	register pt_entry_t	*cpte;
	int			pfn;
	int			num_removed, num_unwired;
	register int		pai;
	register vm_offset_t	pa;

#if	DEBUG_PTE_PAGE
	if (pmap != kernel_pmap)
		ptep_check(get_pte_page(spte));
#endif	DEBUG_PTE_PAGE
	num_removed = 0;
	num_unwired = 0;

	for (cpte = spte; cpte < epte;
	     cpte += ptes_per_vm_page, va += PAGE_SIZE) {
	    pfn = cpte->pfn;
	    if (pfn == 0)
		continue;

	    num_removed++;
	    if (cpte->wired)
		num_unwired++;
	    pa = m88k_ptob(pfn);	/* Physical Address */
	    if (pa < vm_first_phys || pa >= vm_last_phys) {
		/*
		 *	Outside range of managed physical memory.
		 */
		continue;
	    }

	    pai = pa_index(pa);
	    LOCK_PVH(pai);

	    /*
	     *	Get the modify bits.
	     */
	    {
		register int		i;
		register pt_entry_t	*lpte;

		i = ptes_per_vm_page;
		lpte = cpte;
		do {
		    if (lpte->modify) {
			pmap_modify_list[pai] = 1;
			break;
		    }
		    lpte++;
		} while (--i > 0);
	    }

	    /*
	     *	Remove the mapping from the pvlist for
	     *	this physical page.
	     */
	    {
		register pv_entry_t	pv_h, prev, cur;

		pv_h = pai_to_pvh(pai);
		if (pv_h->pmap == PMAP_NULL) {
		    panic("pmap_remove: null pv_list!");
		}
		if (pv_h->va == va && pv_h->pmap == pmap) {
		    /*
		     * Header is the pv_entry.  Copy the next one
		     * to header and free the next one (we cannot
		     * free the header)
		     */
		    cur = pv_h->next;
		    if (cur != PV_ENTRY_NULL) {
			*pv_h = *cur;
			zfree(pv_list_zone, (vm_offset_t) cur);
		    }
		    else {
			pv_h->pmap = PMAP_NULL;
		    }
		}
		else {
		    cur = pv_h;
		    do {
			prev = cur;
			if ((cur = prev->next) == PV_ENTRY_NULL) {
			    panic("pmap-remove: mapping not in pv_list!");
			}
		    } while (cur->va != va || cur->pmap != pmap);
		    prev->next = cur->next;
		    zfree(pv_list_zone, (vm_offset_t) cur);
		}
	    }
	    UNLOCK_PVH(pai);
	}

	/*
	 *	Zero the PTEs to remove the mappings.
	 */

	bzero((caddr_t) spte, (epte-spte)*sizeof(pt_entry_t));

	/*
	 *	Update the counts
	 */
	pmap->stats.resident_count -= num_removed;
	pmap->stats.wired_count -= num_unwired;

	if (pmap != kernel_pmap) {
		register ptep_t	ptep;

		/*
		 *	See whether this pte page can be freed.
		 */
		ptep = get_pte_page(spte);
		ptep->use_count -= num_removed;
		ptep->wired_count -= num_unwired;
#if	DEBUG_PTE_PAGE
		ptep_check(ptep);
#endif	DEBUG_PTE_PAGE
		if (ptep->use_count == 0 && free_if_empty)
			pte_page_free(ptep);
		else if (num_unwired > 0 && ptep->wired_count == 0)
			/* was > 0 before */
			pte_page_unwire(ptep);
	}
}

/*
 *	Remove the given range of addresses
 *	from the specified map.
 *
 *	It is assumed that the start and end are properly
 *	rounded to the page size.
 */

void pmap_remove(map, s, e)
	pmap_t		map;
	vm_offset_t	s, e;
{
	register pt_entry_t	*spte, *epte, *lpte, *syspte;
	int			spl;

	if (map == PMAP_NULL)
		return;

	PMAP_READ_LOCK(map, spl);

	/*
	 *	Invalidate the translation buffer first
	 */
	PMAP_UPDATE_TLBS(map, s, e);

	if (map == kernel_pmap) {
		/*
		 *	Kernel pmap is resident - can do entire
		 *	range at once.
		 */
		spte = pmap_syspte(s);
		epte = spte + m88k_btop(e - s);
		pmap_remove_range(map, s, spte, epte, TRUE);
	}
	else {
		/*
		 *	User pmap.  Break into p0 and p1,
		 *	then check each pte page.
		 */
	    {
		register vm_offset_t p0end, p0start;

		p0end = m88k_ptob(map->p0lr);
		if (s < p0end) {
			/*
			 *	Range includes P0.
			 */

			if (e < p0end)
				p0end = e;
			spte = pmap_p0pte(map, s);
			epte = spte + m88k_btop(p0end - s);
			lpte = (pt_entry_t *)m88k_round_page(spte+1);
			syspte = pmap_syspte(spte);
			p0start = s;
			while (spte < epte) {
				if (lpte > epte)
					lpte = epte;
				if (syspte->valid)
					pmap_remove_range(map, p0start,
						spte, lpte, TRUE);
				p0start += m88k_ptob(lpte-spte);
				spte = lpte;
				lpte += M88K_PGBYTES/sizeof(pt_entry_t);
				syspte++;
			}
		}
	    }
	    {
		register vm_offset_t p1start;

		p1start = m88k_ptob(P0PAGES + map->p1lr);
		if (e > p1start) {
			/*
			 *	Range includes P1.
			 */
			if (s > p1start)
				p1start = s;
			spte = pmap_p1pte(map, p1start);
			epte = spte + m88k_btop(e - p1start);
			lpte = (pt_entry_t *)m88k_round_page(spte+1);
			syspte = pmap_syspte(spte);
			while (spte < epte) {
				if (lpte > epte)
					lpte = epte;
				if (syspte->valid)
					pmap_remove_range(map, p1start,
						spte, lpte, TRUE);
				p1start += m88k_ptob(lpte-spte);
				spte = lpte;
				lpte += M88K_PGBYTES/sizeof(pt_entry_t);
				syspte++;
			}
		}
	    }
	}

	PMAP_READ_UNLOCK(map, spl);
}

/*
 *	Routine:	pmap_remove_all
 *	Function:
 *		Removes this physical page from
 *		all physical maps in which it resides.
 *		Reflects back modify bits to the pager.
 */
static void pmap_remove_all(phys)
	vm_offset_t	phys;
{
	register pv_entry_t	pv_h;
	register pt_entry_t	*pte;
	int			pai;
	register vm_offset_t	va;
	register pmap_t		pmap;
	int			spl;

	if (phys < vm_first_phys || phys >= vm_last_phys) {
	    /*
	     *	Not a managed page.
	     */
	    return;
	}

	/*
	 *	Lock the pmap system first, since we will be changing
	 *	several pmaps.
	 */

	PMAP_WRITE_LOCK(spl);

	/*
	 *	Walk down PV list, removing all mappings.
	 *	We have to do the same work as in pmap_remove_pte_page
	 *	since that routine locks the pv_head.  We don't have
	 *	to lock the pv_head, since we have the entire pmap system.
	 */
	pai = pa_index(phys);
	pv_h = pai_to_pvh(pai);

	while ( (pmap = pv_h->pmap) != PMAP_NULL) {
	    va = pv_h->va;

	    /*
	     *	Lock the pmap to block pmap_extract and similar routines.
	     */
	    simple_lock(&pmap->lock);

	    pte = pmap_pte(pmap, va);

	    if (!pte_page_valid(pte))
		panic("pmap_remove_all: pte invalid");

	    if (m88k_ptob(pte->pfn) != phys) {
		panic("pmap_remove_all: pte doesn't point to page!");
	    }

	    pmap->stats.resident_count--;
	    if (pte->wired) {
		panic("pmap_remove_all removing a wired page");
	    }

	    /*
	     *	Tell CPU using pmap to invalidate its TLB
	     */
	    PMAP_UPDATE_TLBS(pmap, va, va + PAGE_SIZE);

	    /*
	     *	Remove the pv_entry from the pv_list.
	     */
	    {
		register pv_entry_t	cur;

		if ((cur = pv_h->next) != PV_ENTRY_NULL) {
		    *pv_h = *cur;
		    zfree(pv_list_zone, (vm_offset_t) cur);
		}
		else
		    pv_h->pmap = PMAP_NULL;
	    }

	    /*
	     *	Remove the mapping, collecting any modify bits.
	     */
	    {
		register pt_entry_t	*lpte;
		register int		i;

		i = ptes_per_vm_page;
		lpte = pte;
		do {
		    if (lpte->modify) {
			pmap_modify_list[pai] = 1;
		    }
		    *(int *)lpte++ = 0;
		} while (--i > 0);
	    }

	    /*
	     *	Free the pte page if it is no longer needed.
	     */
	    if (pmap != kernel_pmap) {
		register ptep_t	ptep;

		ptep = get_pte_page(pte);
		if (--ptep->use_count == 0)
		    	pte_page_free(ptep);
#if	DEBUG_PTE_PAGE
		else
		    	ptep_check(ptep);
#endif	DEBUG_PTE_PAGE
	    }
	    simple_unlock(&pmap->lock);
	}

	PMAP_WRITE_UNLOCK(spl);
}

boolean_t pmap_verify_free(phys)
	vm_offset_t	phys;
{
	pv_entry_t	pv_h;
	int		pai;
	int		spl;
	boolean_t	result;

	if (!pmap_initialized)
		return(TRUE);

	if (phys < vm_first_phys || phys >= vm_last_phys)
		return(FALSE);

	PMAP_WRITE_LOCK(spl);

	pai = pa_index(phys);
	pv_h = pai_to_pvh(pai);

	result = (pv_h->pmap == PMAP_NULL);
	PMAP_WRITE_UNLOCK(spl);

	return(result);
}

/*
 *	Routine:	pmap_copy_on_write
 *	Function:
 *		Remove write privileges from all
 *		physical maps for this physical page.
 */
static void pmap_copy_on_write(phys)
	vm_offset_t	phys;
{
	register pv_entry_t	pv_e;
	register pt_entry_t	*pte;
	register int		i;
	int			spl;

	/*
	 *	Lock the entire pmap system, since we may be changing
	 *	several maps.
	 */
	PMAP_WRITE_LOCK(spl);

	pv_e = pai_to_pvh(pa_index(phys));
	if (pv_e->pmap == PMAP_NULL) {
		PMAP_WRITE_UNLOCK(spl);
		return;		/* no mappings */
	}

	/*
	 *	Run down the list of mappings to this physical page,
	 *	disabling write privileges on each one.
	 */

	while (pv_e != PV_ENTRY_NULL) {
		pmap_t		pmap;
		vm_offset_t	va;

		pmap = pv_e->pmap;
		va = pv_e->va;

		simple_lock(&pmap->lock);

		pte = pmap_pte(pmap, va);

		if (!pte_page_valid(pte))
			panic("pmap_copy_on_write: invalid pte");

		/*
		 *	Ask cpus using pmap to invalidate their TLBs
		 */
		PMAP_UPDATE_TLBS(pmap, va, va + PAGE_SIZE);

		/*
		 *  In old code, M88K_PTE_WP was set when M88K_PTE_WP was 
		 *  already set.
		 *  This is obvious mistake. We always set M88K_PROT_SET.
		 */
		i = ptes_per_vm_page;
		do {
		    *(int *)pte |= M88K_PTE_WP;
		    if(pte->modify)
			pmap_modify_list[pa_index(phys)] = 1;
		    pte++;
		} while (--i > 0);
		simple_unlock(&pmap->lock);

		pv_e = pv_e->next;
	}

	PMAP_WRITE_UNLOCK(spl);
}

/*
 *	Set the physical protection on the
 *	specified range of this map as requested.
 */
void pmap_protect(map, s, e, prot)
	pmap_t		map;
	vm_offset_t	s, e;
	vm_prot_t	prot;
{
	register pt_entry_t	*spte, *epte, *lpte;
	register int 		vp;
	vm_offset_t	p0end, p1start;
	int		spl;

	if (map == PMAP_NULL)
		return;
	vp = m88k_protection(map, prot);

	SPLVM(spl);
	simple_lock(&map->lock);

	/*
	 *	Invalidate the translation buffer first
	 */
	PMAP_UPDATE_TLBS(map, s, e);

	if (map == kernel_pmap) {
		/*
		 *	Kernel pmap is resident - can do entire
		 *	range at once.
		 */
		spte = pmap_syspte(s);
		epte = spte + m88k_btop(e - s);
		while (spte < epte) {
		    *(int *)spte |= (*(int *)vp & M88K_PROT_BITS);
		    spte++;
		}
	}
	else {
		/*
		 *	User pmap.  Break into p0 and p1,
		 *	then check each pte page.
		 */
		p0end = m88k_ptob(map->p0lr);
		if (s < p0end) {
			/*
			 *	Range includes P0.
			 */
			if (e < p0end)
				p0end = e;
			spte = pmap_p0pte(map, s);
			epte = spte + m88k_btop(p0end - s);
			while (spte < epte) {
				lpte = (pt_entry_t *)m88k_round_page(spte+1);
				if (lpte > epte)
					lpte = epte;
				if (pte_page_valid(spte)) {
					for (; spte < lpte; spte++)
					    *(int *)spte |= (*(int *)vp & M88K_PROT_BITS);
				}
				else {
					spte = lpte;
				}
			}
		}
		p1start = m88k_ptob(P0PAGES + map->p1lr);
		if (e > p1start) {
			/*
			 *	Range includes P1.
			 */
			if (s > p1start)
				p1start = s;
			spte = pmap_p1pte(map, p1start);
			epte = spte + m88k_btop(e - p1start);
			while (spte < epte) {
				lpte = (pt_entry_t *)m88k_round_page(spte+1);
				if (lpte > epte)
					lpte = epte;
				if (pte_page_valid(spte)) {
					for (; spte < lpte; spte++)
					        M88K_PROT_SET(spte, vp);
				}
				else {
					spte = lpte;
				}
			}
		}
	}

	simple_unlock(&map->lock);
	SPLX(spl);
}

/*
 *	Insert the given physical page (p) at
 *	the specified virtual address (v) in the
 *	target physical map with the protection requested.
 *
 *	If specified, the page will be wired down, meaning
 *	that the related pte can not be reclaimed.
 *
 *	NB:  This is the only routine which MAY NOT lazy-evaluate
 *	or lose information.  That is, this routine must actually
 *	insert this page into the given map NOW.
 */
void pmap_enter(pmap, v, pa, prot, wired)
	register pmap_t		pmap;
	vm_offset_t		v;
	register vm_offset_t	pa;
	vm_prot_t		prot;
	boolean_t		wired;
{
	register pt_entry_t	*pte;
	pv_entry_t		pv_e;
	int			spl, vp;
	ptep_t			ptep;
	vm_offset_t		old_pa;

	if (pmap == PMAP_NULL)
		return;

	vp = m88k_protection(pmap, prot);
	/*
	 *  When physcal address is bigger than MAXPHYSMEM,
	 *  the specify address is I/O (like frame memory).
	 */
	if (pa > MAXPHYSMEM)
	    vp = (vp & ~(M88K_PTE_GL | M88K_PTE_WT)) | M88K_PTE_CI;

	/*
	 *	Must allocate a new pvlist entry while we're unlocked;
	 *	zalloc may cause pageout (which will lock the pmap system).
	 *	If we determine we need a pvlist entry, we will unlock
	 *	and allocate one.  Then we will retry, throwing away
	 *	the allocated entry later (if we no longer need it).
	 */
	pv_e = PV_ENTRY_NULL;
Retry:

	PMAP_READ_LOCK(pmap, spl);

	/*
	 *	Expand pmap to include this pte.  Assume that
	 *	pmap is always expanded to include enough 
	 *	pages to map one VM page.
	 */

	while ((pte = pmap_pte(pmap, v)) == PT_ENTRY_NULL) {
		/*
		 *	Must unlock to expand the pmap.
		 */
		PMAP_READ_UNLOCK(pmap, spl);

		user_pt_allocate(pmap, v);

		PMAP_READ_LOCK(pmap, spl);
	}

	/*
	 *	Allocate a physical page for this pte if there isn't
	 *	one already
	 */
	if (pmap == kernel_pmap)
		ptep = PTE_PAGE_NULL;
	else if (pte_page_valid(pte))
		ptep = get_pte_page(pte);
	else {
	    /*
	     *	Try to get a free page-table page.
	     */
	    ptep = pte_free_page_alloc(pmap, v);
	    if (ptep == PTE_PAGE_NULL) {
		/*
		 *	No free pte pages.  Must get the write lock
		 *	on the pmap system to grab one from another pmap.
		 */
		PMAP_READ_UNLOCK(pmap, spl);

		/*
		 *	It's possible that the pte may have gone away...
		 */
		PMAP_WRITE_LOCK(spl);
		pte = pmap_pte(pmap, v);
		if (pte == PT_ENTRY_NULL) {
		    /*
		     *	If it has, back out and try again
		     */
		    PMAP_WRITE_UNLOCK(spl);
		    goto Retry;
		}
		/*
		 * Next follow 3 lines added
		 * Becouse the rase condition caused for Multi-cpus
		 * in PMAP_WRITE_LOCK, we should check again for ptep.
		 */
		if (pte_page_valid(pte))
		    ptep = get_pte_page(pte);
		else
		    ptep = pte_page_alloc(pmap, v);
		PMAP_WRITE_TO_READ_LOCK(pmap);
	    }
	}

#if	DEBUG_PTE_PAGE
	ptep_check(ptep);
#endif	DEBUG_PTE_PAGE
	/*
	 *	Special case if the physical page is already mapped
	 *	at this address.
	 */
	old_pa = m88k_ptob(pte->pfn);
	if (old_pa == pa && pa != 0) {
	    /*
	     *	May be changing its wired attribute or protection
	     */
		
	    if (wired && !pte->wired) {
		pmap->stats.wired_count++;
		if (ptep && ptep->wired_count++ == 0)
		    pte_page_wire(ptep);
	    }
	    else if (!wired && pte->wired) {
		pmap->stats.wired_count--;
		if (ptep && --ptep->wired_count > 0)
		    pte_page_unwire(ptep);
	    }
	    {
		register int template, i;

		template = M88K_PTE_VALID | vp | m88k_trunc_page(pa);
		if (wired)
		    template |= M88K_PTE_WIRED;
		PMAP_UPDATE_TLBS(pmap, v, v + PAGE_SIZE);
		i = ptes_per_vm_page;
		do {
		if (pte->modify)
		    template |= M88K_PTE_MOD;
		*(int *)pte++ = template;
		template += M88K_PGBYTES;
		} while (--i > 0);
	    }
#if	DEBUG_PTE_PAGE
	    ptep_check(ptep);
#endif	DEBUG_PTE_PAGE
	}
	else {

	    /*
	     *	Remove old mapping from the PV list if necessary.
	     */
	    if (old_pa != (vm_offset_t) 0) {
		/*
		 *	Invalidate the translation buffer,
		 *	then remove the mapping.
		 */
		PMAP_UPDATE_TLBS(pmap, v, v + PAGE_SIZE);

		/*
		 *	Don't free the pte page if removing last
		 *	mapping - we will immediately replace it.
		 */
		pmap_remove_range(pmap, v, pte,
				pte + ptes_per_vm_page, FALSE);
#if	DEBUG_PTE_PAGE
		ptep_check(ptep);
#endif	DEBUG_PTE_PAGE
	    }

	    if (pa >= vm_first_phys && pa < vm_last_phys) {

		/*
		 *	Enter the mapping in the PV list for this
		 *	physical page.
		 */
		register int		pai;
		register pv_entry_t	pv_h;

		pai = pa_index(pa);
		LOCK_PVH(pai);
		pv_h = pai_to_pvh(pai);

		if (pv_h->pmap == PMAP_NULL) {
		    /*
		     *	No mappings yet
		     */
		    pv_h->va = v;
		    pv_h->pmap = pmap;
		    pv_h->next = PV_ENTRY_NULL;
		}
		else {
#if	DEBUG
		    {
			/* check that this mapping is not already there */
			pv_entry_t	e = pv_h;
			while (e != PV_ENTRY_NULL) {
			    if (e->pmap == pmap && e->va == v)
				panic("pmap_enter: already in pv_list");
			    e = e->next;
			}
		    }
#endif	DEBUG
		    
		    /*
		     *	Add new pv_entry after header.
		     */
                    if (pv_e == PV_ENTRY_NULL) {
                        UNLOCK_PVH(pai);
                        PMAP_READ_UNLOCK(pmap, spl);
                        /*
                         * This isn't correct either: if zalloc blocks in
                         * kmem_alloc, we return with our CPU in the active
                         * set but ipl raised.  We then can deadlock
                         * grabbing the zone lock.
                         */
                        SPLVM(spl);		/* 91/1/17 XXX */
			pv_e = (pv_entry_t) zalloc(pv_list_zone);
                        /* ZALLOC(pv_list_zone, pv_e, pv_entry_t); */
                        SPLX(spl);		/* 91/1/17 XXX */
			goto Retry;
		    }
		    pv_e->va = v;
		    pv_e->pmap = pmap;
		    pv_e->next = pv_h->next;
		    pv_h->next = pv_e;
		    /*
		     *	Remember that we used the pvlist entry.
		     */
		    pv_e = PV_ENTRY_NULL;
		}
		UNLOCK_PVH(pai);
	    }

	    /*
	     *	And count the mapping.
	     */

	    pmap->stats.resident_count++;
	    if (ptep)
		ptep->use_count++;
	    if (wired) {
		pmap->stats.wired_count++;
		if (ptep && ptep->wired_count++ == 0)
		    pte_page_wire(ptep);
	    }

	    {
		/*
		 *	Build a template to speed up entering -
		 *	only the pfn changes.
		 */
		register int template, i;

	        template = M88K_PTE_VALID | vp | m88k_trunc_page(pa);
	    if (wired)
		template |= M88K_PTE_WIRED;
		i = ptes_per_vm_page;
		do {
			*(int *)pte++ = template;
			template += M88K_PGBYTES;
		} while (--i > 0);
	    }
	}
#if	DEBUG_PTE_PAGE
	ptep_check(ptep);
#endif	DEBUG_PTE_PAGE

	PMAP_READ_UNLOCK(pmap, spl);

        if (pv_e != PV_ENTRY_NULL) {
                SPLVM(spl);
                zfree(pv_list_zone, (vm_offset_t) pv_e);
                SPLX(spl);
	    }
}

/*
 *	Routine:	pmap_change_wiring
 *	Function:	Change the wiring attribute for a map/virtual-address
 *			pair.
 *	In/out conditions:
 *			The mapping must already exist in the pmap.
 */
void pmap_change_wiring(map, v, wired)
	register pmap_t	map;
	vm_offset_t	v;
	boolean_t	wired;
{
	register pt_entry_t	*pte;
	register int		i;
	int			spl;
	ptep_t			ptep;

	/*
	 *	We must grab the pmap system lock because we may
	 *	change a pte_page queue.
	 */
	PMAP_READ_LOCK(map, spl);

	if ((pte = pmap_pte(map, v)) == PT_ENTRY_NULL)
		panic("pmap_change_wiring: pte missing");

	if (map == kernel_pmap)
		ptep = PTE_PAGE_NULL;
	else if (!pte_page_valid(pte))
		panic("pmap_change_wiring: pte invalid");
	else
		ptep = get_pte_page(pte);

	if (wired && !pte->wired) {
		/*
		 *	wiring down mapping
		 */
		map->stats.wired_count++;
		if (ptep && ptep->wired_count++ == 0)
			pte_page_wire(ptep);
	}
	else if (!wired && pte->wired) {
		/*
		 *	unwiring mapping
		 */
		map->stats.wired_count--;
		if (ptep && --ptep->wired_count == 0)
			pte_page_unwire(ptep);
	}

	i = ptes_per_vm_page;
	do {
		(pte++)->wired = wired;
	} while (--i > 0);

	PMAP_READ_UNLOCK(map, spl);
}

/*
 *	Routine:	pmap_extract
 *	Function:
 *		Extract the physical page address associated
 *		with the given map/virtual_address pair.
 */

vm_offset_t pmap_extract(pmap, va)
	register pmap_t	pmap;
	vm_offset_t	va;
{
	register pt_entry_t	*pte;
	register vm_offset_t	pa;
	int			spl;

	SPLVM(spl);
	simple_lock(&pmap->lock);
	if ((pte = pmap_pte(pmap, va)) == PT_ENTRY_NULL)
		pa = (vm_offset_t) 0;
	else if (pmap != kernel_pmap && !pte_page_valid(pte))
		pa = (vm_offset_t) 0;
	else
		pa = m88k_ptob(pte->pfn);
	if (pa)
		pa |= va & (M88K_PGBYTES-1);	/* offset within page */
	simple_unlock(&pmap->lock);
	SPLX(spl);
	return(pa);
}


/*
 *      Routine:        pmap_extract_unlocked
 *      Function:
 *               same as pmap_extract except ignores all locks;
 *               useful in the debugger.
 *
 */
vm_offset_t pmap_extract_unlocked(pmap, va)
	register pmap_t	pmap;
	vm_offset_t	va;
{
	register pt_entry_t	*pte;
	register vm_offset_t	pa;

	if ((pte = pmap_pte(pmap, va)) == PT_ENTRY_NULL)
		pa = (vm_offset_t) 0;
	else if (pmap != kernel_pmap && !pte_page_valid(pte))
		pa = (vm_offset_t) 0;
	else
		pa = m88k_ptob(pte->pfn);
	if (pa)
		pa |= va & (M88K_PGBYTES-1);	/* offset within page */
	return(pa);
}


/*
 *	Copy the range specified by src_addr/len
 *	from the source map to the range dst_addr/len
 *	in the destination map.
 *
 *	This routine is only advisory and need not do anything.
 */
void pmap_copy(dst_pmap, src_pmap, dst_addr, len, src_addr)
	pmap_t		dst_pmap;
	pmap_t		src_pmap;
	vm_offset_t	dst_addr;
	vm_size_t	len;
	vm_offset_t	src_addr;
{
}

#ifdef unused
/*
 *	Require that all active physical maps contain no
 *	incorrect entries NOW.  [This update includes
 *	forcing updates of any address map caching.]
 *
 *	Generally used to insure that a thread about
 *	to run will see a semantically correct world.
 */
static void pmap_update()
{
	FLUSH_TLB_ALL();
}
#endif

/*
 *	Routine:	pmap_collect
 *	Function:
 *		Garbage collects the physical map system for
 *		pages which are no longer used.
 *		Success need not be guaranteed -- that is, there
 *		may well be pages which are not referenced, but
 *		others may be collected.
 *	Usage:
 *		Called by the pageout daemon when pages are scarce.
 */
void pmap_collect(p)
	pmap_t		p;
{
	/* does nothing */
}

/*
 *	Routine:	pmap_activate
 *	Function:
 *		Binds the given physical map to the given
 *		processor, and returns a hardware map description.
 */
void pmap_activate(my_pmap, th, my_cpu)
	register pmap_t	my_pmap;
	thread_t	th;
	int		my_cpu;
{
	PMAP_ACTIVATE(my_pmap, th, my_cpu);
}


#ifdef unused
/*
 *	Routine:	pmap_deactivate
 *	Function:
 *		Indicates that the given physical map is no longer
 *		in use on the specified processor.  (This is a macro
 *		in pmap.h)
 */
static void pmap_deactivate(pmap, th, which_cpu)
	pmap_t		pmap;
	thread_t	th;
	int		which_cpu;
{
	PMAP_DEACTIVATE(pmap, th, which_cpu);
}
#endif

/*
 *	Routine:	pmap_kernel
 *	Function:
 *		Returns the physical map handle for the kernel.
 */
pmap_t pmap_kernel()
{
    	return (kernel_pmap);
}

/*
 *	pmap_zero_page zeros the specified (machine independent)
 *	page.
 */
void
pmap_zero_page(phys)
    vm_offset_t phys;
{
    register int i;
    
    i = ptes_per_vm_page;
    phys = m88k_trunc_page(phys);
   
    while (i--) {
	zero_page(phys);
	phys += M88K_PGBYTES;
    }
}

/*
 *	pmap_copy_page
 *
 *	Copy one machine independent page of physical memory to another.
 */
void
pmap_copy_page(src_addr, dst_addr)
vm_offset_t src_addr, dst_addr;
{
    register int     i;
    
    i = PAGE_SIZE / M88K_PGBYTES;
    
    while (i--) {
	copy_page(m88k_trunc_page(src_addr), m88k_trunc_page(dst_addr));
	src_addr += M88K_PGBYTES;
	dst_addr += M88K_PGBYTES;
    }
}

/*
 *	Routine:	pmap_pageable
 *	Function:
 *		Make the specified pages (by pmap, offset)
 *		pageable (or not) as requested.
 *
 *		A page which is not pageable may not take
 *		a fault; therefore, its page table entry
 *		must remain valid for the duration.
 *
 *		This routine is merely advisory; pmap_enter
 *		will specify that these pages are to be wired
 *		down (or not) as appropriate.
 */
void
pmap_pageable(pmap, start, end, pageable)
	pmap_t		pmap;
	vm_offset_t	start;
	vm_offset_t	end;
	boolean_t	pageable;
{
}

#ifdef unused
/*
 *	pmap_redzone
 *
 *	Give the kernel read-only access to the specified address.
 *	This is used to detect stack overflows.  It is assumed that the
 *	address specified is the last possible kernel stack address.
 *	Therefore, we round up to the nearest machine dependent page.
 */
static void pmap_redzone(pmap, addr)
	pmap_t		pmap;
	vm_offset_t	addr;
{
	pt_entry_t	*pte;
	int		spl;

	addr = m88k_round_page(addr);
	SPLVM(spl);
	simple_lock(&pmap->lock);
	if ((pte = pmap_pte(pmap, addr)) != PT_ENTRY_NULL)
		*(int *)pte |= M88K_PTE_KR;
	FLUSH_TLB_PER_PAGE((caddr_t)addr);/*Nishizakikun*/
	simple_unlock(&pmap->lock);
	SPLX(spl);
}
#endif

/*
 *	Clear the modify bits on the specified physical page.
 *
 *	XXX this does not yet work if the physical page is in
 *	    any maps
 */

void pmap_clear_modify(phys)
	register vm_offset_t	phys;
{
	if (phys >= vm_first_phys && phys < vm_last_phys)
	    pmap_modify_list[pa_index(phys)] = 0;
}

/*
 *	pmap_is_modified:
 *
 *	Return whether or not the specified physical page is modified
 *	by any physical maps.
 *
 *	XXX this does not yet return complete information if the page
 *	    is in any pmaps - pmap_clear_reference should have just
 *	    been called
 *	    
 */

boolean_t pmap_is_modified(phys)
	register vm_offset_t	phys;
{
	if (phys >= vm_first_phys && phys < vm_last_phys)
	    return ((boolean_t)pmap_modify_list[pa_index(phys)]);
	else
	    return (FALSE);
}

/*
 *	pmap_clear_reference:
 *
 *	Clear the reference bit on the specified physical page.
 */

void pmap_clear_reference(phys)
	vm_offset_t	phys;
{
    register pmap_t	pmap;
    register pv_entry_t	pv_e;
    register pt_entry_t *pte;
    register vm_offset_t	va;
    register int	i;
    int			spl;

    if (phys < vm_first_phys || phys >= vm_last_phys)
	/*
	 *	Not a manged page.
	 */
	return;
    PMAP_WRITE_LOCK(spl);

    pv_e = pai_to_pvh(pa_index(phys));
    if (pv_e->pmap == PMAP_NULL) {
	PMAP_WRITE_UNLOCK(spl);
	return;
    }

    while ((pv_e != PV_ENTRY_NULL) && ((pmap = pv_e->pmap) != PMAP_NULL)) {
	va = pv_e->va;

	simple_lock(&pmap->lock);

	pte = pmap_pte(pmap, va);

	if (!pte_page_valid(pte)) {
	    panic("pmap_clear_reference: invalid pte");
	}

	if (m88k_ptob(pte->pfn) != phys) {
	    panic("pmap_clear_reference: pte doesn't point to page!");
	}

	PMAP_UPDATE_TLBS(pmap, va, va + PAGE_SIZE);

	i = ptes_per_vm_page;
	do {
	    (pte++)->u = 0;
	} while (--i > 0);

	simple_unlock(&pmap->lock);
	pv_e = pv_e->next;
    }
    PMAP_WRITE_UNLOCK(spl);
}

/*
 *	pmap_is_referenced:
 *
 *	Return whether or not the specified physical page is referenced
 *	by any physical maps.
 */

boolean_t pmap_is_referenced(phys)
	vm_offset_t	phys;
{
    register pmap_t	pmap;
    register pv_entry_t	pv_e;
    register pt_entry_t *pte;
    register vm_offset_t	va;
    register int	i;
    int			spl;

    if (phys < vm_first_phys || phys >= vm_last_phys)
	/*
	 *	Not a manged page.
	 */
	return(FALSE);
    PMAP_WRITE_LOCK(spl);

    pv_e = pai_to_pvh(pa_index(phys));
    if (pv_e->pmap == PMAP_NULL) {
	PMAP_WRITE_UNLOCK(spl);
	return(FALSE);
    }

    while ((pv_e != PV_ENTRY_NULL) && ((pmap = pv_e->pmap) != PMAP_NULL)) {
	va = pv_e->va;

	simple_lock(&pmap->lock);

	pte = pmap_pte(pmap, va);

	if (!pte_page_valid(pte)) {
	    panic("pmap_is_referenced: invalid pte");
	}

	if (m88k_ptob(pte->pfn) != phys) {
	    panic("pmap_is_referenced: pte doesn't point to page!");
	}

	i = ptes_per_vm_page;
	do {
	    if ((pte++)->u) {
		simple_unlock(&pmap->lock);
		PMAP_WRITE_UNLOCK(spl);
		return(TRUE);
	    }
	} while (--i > 0);

	simple_unlock(&pmap->lock);
	pv_e = pv_e->next;
    }
    PMAP_WRITE_UNLOCK(spl);
    return(FALSE);
}

/*
*	    TLB Coherence Code (TLB "shootdown" code)
* 
* Threads that belong to the same task share the same address space and
* hence share a pmap.  However, they  may run on distinct cpus and thus
* have distinct TLBs that cache page table entries. In order to guarantee
* the TLBs are consistent, whenever a pmap is changed, all threads that
* are active in that pmap must have their TLB updated. To keep track of
* this information, the set of cpus that are currently using a pmap is
* maintained within each pmap structure (cpus_using). Pmap_activate() and
* pmap_deactivate add and remove, respectively, a cpu from this set.
* Since the TLBs are not addressable over the bus, each processor must
* flush its own TLB; a processor that needs to invalidate another TLB
* needs to interrupt the processor that owns that TLB to signal the
* update.
* 
* Whenever a pmap is updated, the lock on that pmap is locked, and all
* cpus using the pmap are signaled to invalidate. All threads that need
* to activate a pmap must wait for the lock to clear to await any updates
* in progress before using the pmap. They must ACQUIRE the lock to add
* their cpu to the cpus_using set. An implicit assumption made
* throughout the TLB code is that all kernel code that runs at or higher
* than splvm blocks out update interrupts, and that such code does not
* touch pageable pages.
* 
* A shootdown interrupt serves another function besides signaling a
* processor to invalidate. The interrupt routine (pmap_update_interrupt)
* waits for the both the pmap lock (and the kernel pmap lock) to clear,
* preventing user code from making implicit pmap updates while the
* sending processor is performing its update. (This could happen via a
* user data write reference that turns on the modify bit in the page
* table). It must wait for any kernel updates that may have started
* concurrently with a user pmap update because the IPC code
* changes mappings.
* Spinning on the VALUES of the locks is sufficient (rather than
* having to acquire the locks) because any updates that occur subsequent
* to finding the lock unlocked will be signaled via another interrupt.
* (This assumes the interrupt is cleared before the low level interrupt code 
* calls pmap_update_interrupt()). 
* 
* The signaling processor must wait for any implicit updates in progress
* to terminate before continuing with its update. Thus it must wait for an
* acknowledgement of the interrupt from each processor for which such
* references could be made. For maintaining this information, a set
* cpus_active is used. A cpu is in this set if and only if it can 
* use a pmap. When pmap_update_interrupt() is entered, a cpu is removed from
* this set; when all such cpus are removed, it is safe to update.
* 
* Before attempting to acquire the update lock on a pmap, a cpu (A) must
* be at least at the priority of the interprocessor interrupt
* (splip<=splvm). Otherwise, A could grab a lock and be interrupted by a
* kernel update; it would spin forever in pmap_update_interrupt() trying
* to acquire the user pmap lock it had already acquired. Furthermore A
* must remove itself from cpus_active.  Otherwise, another cpu holding
* the lock (B) could be in the process of sending an update signal to A,
* and thus be waiting for A to remove itself from cpus_active. If A is
* spinning on the lock at priority this will never happen and a deadlock
* will result.
*/

int shotclock=0;
simple_lock_data_t shotlock=0;

int shootdowns[NCPUS] = {0,0,0,0};

/* find leftmost set bit */
static unsigned fls(x)
register unsigned x;
{
  if (x==0) return 0;
  asm volatile ("ff1	%0, %0" : "=r" (x) : "0" (x));
  return x+1;
}
    
/*
 *	Signal another CPU that it must flush its TLB
 *
 *      If a cpu is idle. we don't shoot it down, merely
 *      queue its invalidations for user pmap events.
 *      However, all shootdowns on the kernel pmap are
 *      immediately propogated to everyone, as interrupts
 *      can occur while a cpu is idle and use kernel VM
 *      that has been manipulated by different processors. 
 *      This also assumes that interrupt routines NEVER
 *      use user memory.
 *
 */
static void signal_cpus(use_list, pmap, start, end)
	register cpu_set	use_list;
	pmap_t			pmap;
	vm_offset_t		start, end;
{
	register int		which_cpu, j;
	register pmap_update_list_t	update_list_p;
	register pmap_update_item_t	update_item_p;

	shootdowns[cpu_number()]++;

	while ((which_cpu = fls(use_list)) != 0) {
	    which_cpu -= 1;	/* convert to 0 origin */

	    update_list_p = &cpu_update_list[which_cpu];
	    simple_lock(&update_list_p->lock);

	    j = update_list_p->count;
	    if (j >= UPDATE_LIST_SIZE) {
		update_list_p->overflowed = TRUE;
	    }
	    else {
		update_item_p = &update_list_p->item[j];
		update_item_p->pmap  = pmap;
		update_item_p->start = start;
		update_item_p->end   = end;

		update_list_p->count = j+1;
	    }
	    cpu_update_needed[which_cpu] = TRUE;
	    simple_unlock(&update_list_p->lock);

	    /* if its the kernel pmap, ignore cpus_idle */
	    if (((cpus_idle & (1 << which_cpu)) == 0) ||
		(pmap == kernel_pmap))
	      {
		interrupt_processor(which_cpu);
	      }
	    use_list &= ~(1 << which_cpu);
	}
}

void process_pmap_updates(my_pmap)
	register pmap_t		my_pmap;
{
	register int		my_cpu = cpu_number();
	register pmap_update_list_t	update_list_p;
	register int		j;
	register pmap_t		pmap;

	update_list_p = &cpu_update_list[my_cpu];
	simple_lock(&update_list_p->lock);
	if ( update_list_p->overflowed){
	    /* invalidate all PATCs of kernel(supervisor) and user */
	    cmmu_flush_tlb(FLUSH_KERNEL, 0, FLUSH_ALL);
	    cmmu_flush_tlb(FLUSH_USER, 0, FLUSH_ALL);
	    update_list_p->overflowed = FALSE;
	}else{
	    for (j = 0; j < update_list_p->count; j++) {
		pmap = update_list_p->item[j].pmap;
		if (pmap == my_pmap ||
		    pmap == kernel_pmap) {
		    
		    INVALIDATE_TLB(pmap == kernel_pmap,
				   (caddr_t)update_list_p->item[j].start,
				   update_list_p->item[j].end);
		}
	    }
	}
	update_list_p->count = 0;
	cpu_update_needed[my_cpu] = FALSE;
	simple_unlock(&update_list_p->lock);
}

/*
 *	Interrupt routine for TBIA requested from other processor.
 */
void pmap_update_interrupt(my_cpu)
int my_cpu;
{
	register pmap_t		my_pmap;
	int			s;

xpr("_update_interrupt shotclock 0x%x\n",shotclock);

	if ( my_cpu != cpu_number() )
	    printf("(%d)pmap_update_interrupt: Soft.int.para.pass err.(%d)\n",
		   cpu_number(),my_cpu);

	/* This code used to return when the cpu was idle. Interrupts,
	   however make that a bad idea. */

	/*    
	 * find my_pmap
	 */
	if ( current_thread() == THREAD_NULL )
	    my_pmap = kernel_pmap;
	else {
	    my_pmap = current_pmap();
	    if (!pmap_in_use(my_pmap, my_cpu))
		my_pmap = kernel_pmap; 
	}
	s = splvm();
	
	do {
	  /* This is necessary because of a bug in gcc 1.37 */
	  volatile simple_lock_t my_map_lock = &(my_pmap->lock);
	  volatile simple_lock_t kernel_map_lock = &(kernel_pmap->lock);
	  
	    /*
	     *	Indicate that we're not using either user or kernel
	     *	pmap.
	     */

	    i_bit_clear(my_cpu, &cpus_active, &cpus_active_lock);
	    /*
	     *	Wait for any pmap updates in progress, on either user
	     *	or kernel pmap.
	     */
	    while (my_map_lock->lock_data || kernel_map_lock->lock_data)
		continue;

	    process_pmap_updates(my_pmap);

	    i_bit_set(my_cpu, &cpus_active, &cpus_active_lock);

	    if ((my_pmap != kernel_pmap) &&
		(my_pmap->seg_changed & (1 << my_cpu))) {
		pmap_set_user_segment(my_pmap);
	    }
	} while (cpu_update_needed[my_cpu]);
	splx(s);
}

/* 
 *    Routine:    interrupt_processor
 *    Function:
 *        cause an interrupt in the specified cpu
 */
static void interrupt_processor(which_cpu)
int which_cpu;
{
xpr("softcalling cpu %d for shotclock %d\n",which_cpu, shotclock);
    softcall((func_t)pmap_update_interrupt, (caddr_t)which_cpu, which_cpu);
}


/*
 * 
 *      shootdown_wait (pmap, mask_of_cpus_to_wait_for)
 *
 *      wait for a pmap to be freed by the specified bit mask of processors.
 */
static int shootdown_wait(pmap, cpu_mask)
pmap_t pmap;
int cpu_mask;
{
  volatile pmap_t p = pmap;
  cpu_set *active = &cpus_active;
  
xpr("shootdown_wait: pmap 0x%x mask 0x%x cpus_active 0x%x\n",
    pmap, cpu_mask, cpus_active);

  while (1)
    {
      /* determine which cpus we have to worry about */
      int worry = p->cpus_using & cpu_mask & *active; 

      if (!worry)
	break;
    } /* end while 1 */

  return 1;
}

void pmap_update_tlbs(pmap, start, end)
pmap_t pmap;
vm_offset_t start,end;
{
  cpu_set cpu_mask = 1 << cpu_number();
  cpu_set users;
  /* find other cpus using the pmap */
  users = pmap->cpus_using & ~cpu_mask;


  if (users) 
    {
      simple_lock(&shotlock);
      shotclock++;
      simple_unlock(&shotlock);
xpr("_update_tlbs: pmap 0x%x start 0x%x end 0x%x users 0x%x shotclock %d\n",
    pmap, start, end, users,shotclock);
      /* signal them,... */
      signal_cpus(users, pmap, start, end);
      /* ...and, wait for them to finish using the pmap. */
      shootdown_wait(pmap, ~cpu_mask); 
    }
  /* If pmap is in use, invalidate ATCs of this CPU. */\
    if ((pmap)->cpus_using & cpu_mask) 
      {
	INVALIDATE_TLB( (pmap == kernel_pmap), (caddr_t)start, end);
	if ((pmap != kernel_pmap) && (((pmap)-> seg_changed) & cpu_mask))
	  pmap_set_user_segment(pmap);		
      }
  
  return;
}

/*
 *	pmap_page_protect:
 *
 *	Lower the permission for all mappings to a given page.
 */
void	pmap_page_protect(phys, prot)
	vm_offset_t	phys;
	vm_prot_t	prot;
{
	switch (prot) {
		case VM_PROT_READ:
		case VM_PROT_READ|VM_PROT_EXECUTE:
			pmap_copy_on_write(phys);
			break;
		case VM_PROT_ALL:
			break;
		default:
			pmap_remove_all(phys);
			break;
	}
}

void pmap_set_user_segment(pmap)
pmap_t	pmap;
{
    register int n, mask;
    register pt_entry_t	*syspte;
    register segment_d_t	*ste,*sste;
    register struct user_segment_info *usip;
    register int my_cpu = cpu_number();

    if (pmap == kernel_pmap)
	return;

    /*
     * Get user segment info & user segment
     */
    usip = &user_segment_info[my_cpu];

    /*
     * Set new segment map(p0 region)
     */
    ste = sste = 
	(segment_d_t *)&(((struct segment_map *)user_segment_maps)[my_cpu]);
    syspte = pmap_syspte(pmap->p0br);
    if (cache_policy & M88K_PTE_CI)
	mask = ~M88K_PTE_SP;
    else if (cache_policy & M88K_PTE_WT)
	mask = ~(M88K_PTE_SP | M88K_PTE_CI);
    else
	mask = ~(M88K_PTE_SP | M88K_PTE_WT | M88K_PTE_CI);
    for (n = 0; n < pmap->p0lr; n+=(M88K_PGBYTES/sizeof(segment_d_t))) {
	*(int *)ste++ = *(int *)syspte++ & mask;
    }

    /*
     * Clear old segment map(p0 region)
     */
    for (; n < usip->p0lr; n += (M88K_PGBYTES/sizeof(segment_d_t))) {
	*(int *)ste++ = 0;
    }

    /*
     * Set new segment (p1 region)
     */
    ste = &sste[(P0PAGES+P1PAGES)/(M88K_PGBYTES/sizeof(segment_d_t))];
    syspte = pmap_syspte(pmap->p1br + P1PAGES);
    for (n=0; n< P1PAGES - pmap->p1lr; n+=(M88K_PGBYTES/sizeof(segment_d_t))) {
	*(int *)--ste = *(int *)--syspte & mask;
    }

    /*
     * Clear old segment (p1 region)
     */
    for (; n < P1PAGES - usip->p1lr; n += NBPG/sizeof(segment_d_t))
	*(int *)--ste = 0;

    /*
     * Copy user segment info.
     */
    usip->p0lr = pmap->p0lr;
    usip->p1lr = pmap->p1lr;

    /* Clear a seg_changed bit */
    pmap->seg_changed &= ~(1 << my_cpu);

    /*
     * Set CMMU USER AREA DESCRIPTION REGISTER
     */
    cmmu_set_uapr((int)sste | cache_policy | M88K_AD_TE);
    for (n = 0; n < NBATC; n++) {
	cmmu_set_batc_entry(n, 0, pmap->i_batc[n].raw);
	cmmu_set_batc_entry(n, 1, pmap->d_batc[n].raw);
    }
}

#ifdef unused
/*
 *	Set BATC
 */
static void
pmap_set_batc(pmap, data, i, va, pa, super, wt, global, ci, wp, valid)
pmap_t pmap;
register int i;
register vm_offset_t va, pa;
boolean_t data, super, wt, global, ci, wp, valid;
{
    union batc_access *ptr;

    if (i < 0 || i > (NBATC - 1)) {
	panic("pmap_set_batc: illegal batc number\n");
	/* bad number */
	return;
    }
    if (data) {
	ptr = &pmap->d_batc[i];
    } else {
	ptr = &pmap->i_batc[i];
    }

    ptr->cooked.lba = va >> 19;
    ptr->cooked.pba = pa >> 19;
    ptr->cooked.s = super;
    ptr->cooked.wt = wt;
    ptr->cooked.g = global;
    ptr->cooked.ci = ci;
    ptr->cooked.wp = wp;
    ptr->cooked.v = valid;
    cmmu_set_batc_entry(i, data, ptr->raw);
}
#endif

void
pmap_set_kernel_mmu()
{
 	    
    /*
     * Check kernel_segment_map address
     *	we must be bound a page size in segment map.
     */
    assert((int)kernel_segment_map & ~(M88K_PGBYTES-1));

    /*
     * MMU enable. Kernel code is the same address both V and R.
     * kernel_area_descriptor is kernel_segment_map + some flags.
     */
    cmmu_set_sapr(kernel_area_descriptor);

}


/*
 *	User page table space management.
 */
static void user_pt_init()
{
#ifdef  MACH_KERNEL
        if (kmem_alloc_pageable(kernel_map,&user_pt_min, user_pt_total_size)
	    != KERN_SUCCESS)
	  panic("user_pt_init");
#else
	user_pt_min = kmem_alloc_pageable(kernel_map, user_pt_total_size);
#endif
	user_pt_max = user_pt_min + user_pt_total_size;

	user_pt_list.prev = user_pt_list.next = &user_pt_list;
	user_pt_list.end  = user_pt_min;
	user_pt_list.start = user_pt_max;
	user_pt_last_in_use = &user_pt_list;
	user_pt_free_size = user_pt_total_size;
}

/*
 *	Move data by remapping in kernel map.  Assumes that kernel TLB
 *	has already been flushed for old range, and that new range is
 *	unmapped.
 */
static void user_pt_move_ptes(old_start, new_start, size)
	register vm_offset_t	old_start;
	vm_offset_t		new_start;
	register vm_size_t	size;
{
	register pt_entry_t	*opte, *npte;
	register int		count;

	if (old_start == new_start)
	    return;

	opte = pmap_syspte(old_start);
	npte = pmap_syspte(new_start);
	count = m88k_btop(size);

	simple_lock(&kernel_pmap->lock);
	PMAP_UPDATE_TLBS(kernel_pmap, old_start, old_start + size);

	if (new_start < old_start) {
	    /*
	     *	Moving toward lower addresses.
	     *	Move left-to-right.
	     */

	    while (count > 0) {
		*(int *)npte++ = *(int *)opte;
		*(int *)opte++ = 0;
		count--;
	    }
	}
	else {
	    /*
	     *	Moving toward higher addresses.
	     *	Move right-to-left.
	     */
	    opte += count;
	    npte += count;

	    while (count > 0) {
		*(int *)--npte = *(int *)--opte;
		*(int *)opte = 0;
		count--;
	    }
	}
	simple_unlock(&kernel_pmap->lock);
}

/*
 *	Move a user_pt region to a new location.  The pmaps are moved,
 *	the new sizes entered in the region, and the user_pt free space
 *	updated.  Caller will relink the region in the user_pt_list
 *	if necessary.
 */
static void user_pt_move(region, new_start, new_size)
	register pt_region_t	region;
	vm_offset_t		new_start;
	vm_size_t		new_size;
{
	register pmap_t	pmap;
	int		spl;
	vm_offset_t	old_start;
	vm_size_t	old_size;

	old_start = region->start;
	old_size = region->end - old_start;

	region->start = new_start;
	region->end = new_start + new_size;

	user_pt_free_size -= new_size - old_size;

	pmap = region->pmap;

	PMAP_READ_LOCK(pmap, spl);

	if (region == &((a_pmap_t)pmap)->region_0) {
	    /*
	     *	moving region 0
	     */
	    if (old_size != 0) {
		/*
		 *	Invalidate old region in kernel map before we
		 *	move it.  Invalidate region in user map only
		 *	to prevent base and length registers being
		 *	used while the page table is moving, and to
		 *	force the new registers to be loaded (the mapping
		 *	information itself is not changed).
		 */
		PMAP_UPDATE_TLBS(pmap, VM_MIN_ADDRESS, m88k_ptob(pmap->p0lr));

		user_pt_move_ptes(old_start, new_start, old_size);
	    }

	    pmap->p0br = (pt_entry_t *)new_start;
	    pmap->p0lr = new_size / sizeof(pt_entry_t);

	    if (current_thread() != THREAD_NULL &&
		current_pmap() == pmap) {
		pmap_set_user_segment(pmap);
	    }
	}
	else {
	    /*
	     *	moving region 1
	     */
	    if (old_size != 0) {
		PMAP_UPDATE_TLBS(pmap, m88k_ptob(P0PAGES + pmap->p1lr),
				VM_MAX_ADDRESS);

		user_pt_move_ptes(old_start, region->end - old_size, old_size);
	    }

	    pmap->p1br = (pt_entry_t *)(new_start + new_size
					- P1PAGES * sizeof(pt_entry_t));
	    pmap->p1lr = P1PAGES - (new_size / sizeof(pt_entry_t));

	    if (current_thread() != THREAD_NULL &&
		current_pmap() == pmap) {
		pmap_set_user_segment(pmap);
	    }
	}

	PMAP_READ_UNLOCK(pmap, spl);
}

static void user_pt_delete(region)
	register pt_region_t	region;
{
#if	DEBUG_USER_PT
	if (region->next == region)
	    panic("user_pt_delete");
#endif	DEBUG_USER_PT

	if (user_pt_last_in_use->end >= region->start)
	    user_pt_last_in_use = region->prev;
	remque((queue_entry_t)region);
	user_pt_free_size += region->end - region->start;

	/*
	 *	Chain the region to itself so that remque does
	 *	not have to check whether it is on the user_pt_list.
	 */
	region->next = region;
	region->prev = region;
	region->start = 0;
	region->end = 0;
}

/*
 *	Finds a free area of size 'length' in the user_pt map.
 *	Allocates it for 'region' if found.  Skips over 'region'
 *	so that its existing area can be reallocated.
 *
 *	Assumes that region is already in the user_pt_list.
 */
static boolean_t user_pt_find(region, size)
	register pt_region_t	region;
	register vm_size_t	size;
{
	register pt_region_t	entry, next;

	entry = user_pt_last_in_use;
	if (entry == region)
	    entry = entry->prev;

	while (TRUE) {
	    next = entry->next;
	    if (next == region)
		next = next->next;
	    if (next->start - entry->end >= size)
		break;
	    entry = next;
	    if (entry == &user_pt_list)
		return (FALSE);	/* no space */
	}

	/*
	 *	Remove region from its current spot...
	 */
	remque((queue_entry_t)region);

	/*
	 *	Move any contained mappings to the new spot...
	 */
	user_pt_move(region,
		     entry->end,
		     size);

	/*
	 *	and insert it after entry.
	 */
	insque((queue_entry_t)region, (queue_entry_t)entry);

	if (user_pt_last_in_use == entry)
	    user_pt_last_in_use = region;

	return(TRUE);
}

/*
 *	Recycle enough pt_regions to allocate space for a new region.
 *	Marked_region is the region to be reallocated; skip over it
 *	when recycling entries.
 */
void user_pt_collect(marked_region, new_size)
	pt_region_t	marked_region;
	vm_size_t	new_size;
{
	register pt_region_t	entry;
	pt_region_t		first_entry, next_entry;
	int		spl;

	PMAP_WRITE_LOCK(spl);

	/*
	 *	Find a block of regions with no wired-down mappings
	 *	that will give the needed space when freed.
	 */
	first_entry = PT_REGION_NULL;
	for (entry = user_pt_list.next;
	     entry != &user_pt_list;
	     entry = entry->next) {

	    if (entry->pmap->stats.wired_count) {
		/*
		 *	Cannot recycle this region.
		 */
		first_entry = PT_REGION_NULL;
	    }
	    else {
		if (first_entry == PT_REGION_NULL) {
		    first_entry = entry;
		}
		if (entry->next->start - first_entry->prev->end >= new_size)
		    break;
	    }
	}
	if (first_entry == PT_REGION_NULL) {
	    panic("user_pt_collect");
	}

	next_entry = entry->next;

	/*
	 *	Now free up all of the regions that we found.
	 */
	entry = first_entry;
	while (entry != next_entry) {
	    pmap_t	pmap;
	    register vm_offset_t	va;
	    register pt_entry_t	*spte, *epte, *lpte;
	    pt_region_t	next;

	    if (entry == marked_region) {
		entry = entry->next;
		continue;
	    }

	    pmap = entry->pmap;
	    simple_lock(&pmap->lock);

	    if (entry == &((a_pmap_t)pmap)->region_0) {

		PMAP_UPDATE_TLBS(pmap, VM_MIN_ADDRESS,
				m88k_ptob(pmap->p0lr));
		va = (vm_offset_t) 0;
		spte = pmap->p0br;
		epte = spte + pmap->p0lr;

		while (spte < epte) {
		    lpte = (pt_entry_t *)m88k_round_page(spte+1);
		    if (lpte > epte)
			lpte = epte;
		    if (pte_page_valid(spte))
			pmap_remove_range(pmap, va, spte, lpte, TRUE);
		    va += m88k_ptob(lpte - spte);
		    spte = lpte;
		}
		pmap->p0br = (pt_entry_t *)VM_MIN_KERNEL_ADDRESS;
		pmap->p0lr = 0;
	    }
	    else {
		va = m88k_ptob(P0PAGES + pmap->p1lr);
		PMAP_UPDATE_TLBS(pmap, va, VM_MAX_ADDRESS);

		spte = pmap->p1br + pmap->p1lr;
		epte = pmap->p1br + P1PAGES;

		while (spte < epte) {
		    lpte = (pt_entry_t *)m88k_round_page(spte+1);
		    if (lpte > epte)
			lpte = epte;
		    if (pte_page_valid(spte))
			pmap_remove_range(pmap, va, spte, lpte, TRUE);
		    va += m88k_ptob(lpte - spte);
		    spte = lpte;
		}
		pmap->p1br = (pt_entry_t *)VM_MIN_KERNEL_ADDRESS;
		pmap->p1lr = P1PAGES;

	    }

	    /*
	     *	Reload the base registers if the current thread is
	     *	using the pmap.
	     */
	    if (current_thread() != THREAD_NULL && pmap == current_pmap()) {
		pmap_set_user_segment(pmap);
	    }
	    simple_unlock(&pmap->lock);

	    /*
	     *	The kernel mappings for the region have already been freed,
	     *	since all pte pages have been freed.
	     */

	    next = entry->next;
	    user_pt_delete(entry);
	    entry = next;
	}
	PMAP_WRITE_UNLOCK(spl);
}


static void user_pt_destroy(region)
	register pt_region_t	region;
{
	user_pt_lock();

	if (region->end - region->start != 0) {
	    /*
	     *	The kernel mappings for the region have already been freed,
	     *	since all pte pages have been freed.  Just delete the region.
	     */
	    user_pt_delete(region);
	}
#if	DEBUG_USER_PT
	user_pt_check("destroy", PT_REGION_NULL);
#endif	DEBUG_USER_PT
	user_pt_unlock();
}


/*
 *	Routine:	user_pt_allocate
 *
 *	Expands a pmap to be able to map the specified virtual address.
 *
 *	Allocates new virtual memory for the P0 or P1 portion of the
 *	pmap, then re-maps the physical pages that were in the old
 *	pmap to be in the new pmap.
 *
 *	Must be called with the pmap system and the pmap unlocked,
 *	since these must be unlocked to use vm_allocate or vm_deallocate.
 *	Thus it must be called in a loop that checks whether the map
 *	has been expanded enough.
 *	(We won't loop forever, since page tables aren't shrunk.)
 */
static void user_pt_allocate(pmap, v)
	pmap_t		pmap;
	vm_offset_t	v;
{
	register vm_size_t	new_size, old_size;
	register pt_region_t	region;

	if (pmap == kernel_pmap)
		panic("kernel pmap too small");

	user_pt_lock();
USER_PT_LOCK_MARK;/*jeff*/


	if (luna88k_p0(v)) {
	    /*
	     *	Find a new P0 region large enough to map v
	     */
	    register vm_size_t ptsize;

USER_PT_LOCK_MARK;/*jeff*/
	    ptsize = m88k_btop(v + PAGE_SIZE);
	    new_size = round_pte_page(ptsize * sizeof(pt_entry_t));

	    region = &((a_pmap_t)pmap)->region_0;
USER_PT_LOCK_MARK;/*jeff*/
	}
	else {
	    /*
	     *	Same for P1
	     */
	    register vm_size_t ptsize;

USER_PT_LOCK_MARK;/*jeff*/
	    ptsize = m88k_btop(VM_MAX_ADDRESS - v);
	    new_size = round_pte_page(ptsize * sizeof(pt_entry_t));

	    region = &((a_pmap_t)pmap)->region_1;
USER_PT_LOCK_MARK;/*jeff*/
	}

USER_PT_LOCK_MARK;/*jeff*/
	old_size = region->end - region->start;
	if (old_size != 0) {
	    /*
	     *	Region exists.  First see whether it can be extended.
	     */
USER_PT_LOCK_MARK;/*jeff*/
	    if (luna88k_p0(v)) {
		if (region->next->start - region->start >= new_size) {
		    /*
		     *	Extend toward higher addresses.  No need
		     *	to lock pmap, since no existing data changes.
		     */
		    user_pt_free_size -= new_size - old_size;
		    region->end = region->start + new_size;
		    pmap->p0lr = new_size / sizeof(pt_entry_t);
USER_PT_LOCK_MARK;/*jeff*/
		    if (current_thread() != THREAD_NULL &&
			current_pmap() == pmap)
			pmap_set_user_segment(pmap);
USER_PT_LOCK_MARK;/*jeff*/
#if	DEBUG_USER_PT
		    user_pt_check("extend P0", PT_REGION_NULL);
#endif	DEBUG_USER_PT
		    user_pt_unlock();
		    return;
		}
	    }
	    else {
USER_PT_LOCK_MARK;/*jeff*/
		if (region->end - region->prev->end >= new_size) {
		    /*
		     *	Extend toward lower addresses.  No need
		     *	to lock pmap, since no existing data changes.
		     */
USER_PT_LOCK_MARK;/*jeff*/
		    user_pt_free_size -= new_size - old_size;
		    region->start = region->end - new_size;
		    pmap->p1lr = P1PAGES - (new_size / sizeof(pt_entry_t));
		    if (current_thread() != THREAD_NULL &&
			current_pmap() == pmap)
			pmap_set_user_segment(pmap);
#if	DEBUG_USER_PT
		    user_pt_check("extend P1", PT_REGION_NULL);
#endif	DEBUG_USER_PT
USER_PT_LOCK_MARK;/*jeff*/
		    user_pt_unlock();
		    return;
		}
	    }
USER_PT_LOCK_MARK;/*jeff*/
	}

	else {
USER_PT_LOCK_MARK;/*jeff*/
	    /*
	     *	If region is not allocated, link it into the top
	     *	of the region list, with no space allocated,
	     *	to simplify the rest of the code.
	     */
	    enqueue_tail((queue_t)&user_pt_list, (queue_entry_t)region);
USER_PT_LOCK_MARK;/*jeff*/
	    region->start = region->end = user_pt_max;
	}

USER_PT_LOCK_MARK;/*jeff*/
	if (user_pt_free_size >= new_size) {
USER_PT_LOCK_MARK;/*jeff*/
	    /*
	     *	Can allocate region without collecting other regions.
	     */

	    /*
	     *	See if there is enough room to move region
	     */
	    if (region->next->start - region->prev->end >= new_size) {
USER_PT_LOCK_MARK;/*jeff*/
		user_pt_move(region,
			     region->prev->end,
			     new_size);
USER_PT_LOCK_MARK;/*jeff*/

#if	DEBUG_USER_PT
		user_pt_check("extend into free space", PT_REGION_NULL);
#endif	DEBUG_USER_PT
		user_pt_unlock();
		return;
	    }
USER_PT_LOCK_MARK;/*jeff*/

	    /*
	     *	Look for a free area in the user page table.
	     */
	    if (user_pt_find(region, new_size)) {
USER_PT_LOCK_MARK;/*jeff*/
		/*
		 *	Found one; region has been moved here.
		 */
#if	DEBUG_USER_PT
		user_pt_check("find new region", PT_REGION_NULL);
#endif	DEBUG_USER_PT
USER_PT_LOCK_MARK;/*jeff*/
		user_pt_unlock();
		return;
	    }
USER_PT_LOCK_MARK;/*jeff*/

	    /*
	     *	Compress regions between start and current region
	     *	toward bottom of space.
	     */
	    {
		register pt_region_t	entry;

		for (entry = user_pt_list.next;
		     entry != region;
		     entry = entry->next) {
USER_PT_LOCK_MARK;/*jeff*/
		    user_pt_move(entry,
				 entry->prev->end,
				 entry->end - entry->start);
USER_PT_LOCK_MARK;/*jeff*/
		}
USER_PT_LOCK_MARK;/*jeff*/
#if	DEBUG_USER_PT
		user_pt_check("compress to bottom", region);
#endif	DEBUG_USER_PT
	    }

USER_PT_LOCK_MARK;/*jeff*/
	    /*
	     *	See if there is enough room to move region
	     */
	    if (region->next->start - region->prev->end >= new_size) {
USER_PT_LOCK_MARK;/*jeff*/
		user_pt_move(region,
			     region->prev->end,
			     new_size);
USER_PT_LOCK_MARK;/*jeff*/
#if	DEBUG_USER_PT
		user_pt_check("move after compress to bottom", PT_REGION_NULL);
#endif	DEBUG_USER_PT
		user_pt_unlock();
		return;
	    }
USER_PT_LOCK_MARK;/*jeff*/

	    /*
	     *	Compress regions between current region and end
	     *	toward top of space.
	     */
	    {
		register pt_region_t	entry;
		register vm_size_t	cur_size;
USER_PT_LOCK_MARK;/*jeff*/

		for (entry = user_pt_list.prev;
		     entry != region;
		     entry = entry->prev) {
		    cur_size = entry->end - entry->start;
USER_PT_LOCK_MARK;/*jeff*/
		    user_pt_move(entry,
				 entry->next->start - cur_size,
				 cur_size);
USER_PT_LOCK_MARK;/*jeff*/
		}
USER_PT_LOCK_MARK;/*jeff*/
#if	DEBUG_USER_PT
		user_pt_check("compress to top", region);
#endif	DEBUG_USER_PT
	    }

USER_PT_LOCK_MARK;/*jeff*/
	    /*
	     *	See if there is enough room to move region
	     */
	    if (region->next->start - region->prev->end >= new_size) {
USER_PT_LOCK_MARK;/*jeff*/
		user_pt_move(region,
			     region->prev->end,
			     new_size);
USER_PT_LOCK_MARK;/*jeff*/
#if	DEBUG_USER_PT
		user_pt_check("move after compress to top", PT_REGION_NULL);
#endif	DEBUG_USER_PT
		user_pt_unlock();
		return;
	    }
USER_PT_LOCK_MARK;/*jeff*/
	}

USER_PT_LOCK_MARK;/*jeff*/
	/*
	 *	Not enough room even when compressing entries.
	 *	Remove some, marking the current entry so that
	 *	it is not collected but its space can count toward
	 *	free space.
	 */
	user_pt_collect(region, new_size);
USER_PT_LOCK_MARK;/*jeff*/
#if	DEBUG_USER_PT
	user_pt_check("after collect", region);
#endif	DEBUG_USER_PT

	/*
	 *	Look for a free area in the user page table.
	 */
	if (user_pt_find(region, new_size)) {
USER_PT_LOCK_MARK;/*jeff*/
	    /*
	     *	Found one; region has been moved here.
	     */
#if	DEBUG_USER_PT
	    user_pt_check("find after collect", PT_REGION_NULL);
#endif	DEBUG_USER_PT
	    user_pt_unlock();
	    return;
	}
USER_PT_LOCK_MARK;/*jeff*/

	/*
	 *	All out!
	 */
	panic("user_pt_allocate");
}

#if	DEBUG_USER_PT
user_pt_check(msg, exempt_region)
	char	*msg;
	register pt_region_t	exempt_region;
{
	register pt_region_t	entry;
	register vm_size_t	used = 0;
	register pt_entry_t	*spte, *epte;
	register boolean_t	found_last_in_use = FALSE;

	if (user_pt_last_in_use == &user_pt_list)
	    found_last_in_use = TRUE;

	for (entry = user_pt_list.next;
	     entry != &user_pt_list;
	     entry = entry->next) {
	    if (entry->prev->next != entry ||
		entry->end < entry->start ||
		entry->prev->end > entry->start ||
		entry->end > entry->next->start) {
		printf("%s\n", msg);
		panic("user_pt_check: entries out of order");
		}
	    if (entry != exempt_region && entry->end == entry->start) {
		printf("%s\n", msg);
		panic("user_pt_check: zero-length region:");
	    }
	    used += entry->end - entry->start;
	    spte = pmap_syspte(entry->prev->end);
	    epte = pmap_syspte(entry->start);
	    for (;spte < epte; spte++) {
		if (*(int *)spte != 0) {
		    printf("%s\n", msg);
		    panic("user_pt_check: unallocated region in use");
		}
	    }
	    if (user_pt_last_in_use == entry)
		found_last_in_use = TRUE;
	}
	if (!found_last_in_use) {
	    panic("user_pt_check: last_in_use not in list");
	}
	if (user_pt_free_size + used != user_pt_total_size) {
	    printf("%s\n", msg);
	    panic("user_pt_check: free size mismatch");
	}
}
#endif	DEBUG_USER_PT

/* New functions to satisfy rpd */

void pmap_virtual_space(startp, endp)
	vm_offset_t *startp;
	vm_offset_t *endp;
{
	*startp = virtual_avail;
	*endp = virtual_end;
}

unsigned int pmap_free_pages()
{
	return atop(avail_end - avail_next);
}

boolean_t pmap_next_page(addrp)
	vm_offset_t *addrp;
{
	if (avail_next == avail_end)
		return FALSE;

	*addrp = avail_next;
	avail_next += PAGE_SIZE;
	return TRUE;
}
