/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS AS-IS
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * HISTORY
 * $Log:	kmin_dma.c,v $
 * Revision 2.6  92/03/02  18:34:10  rpd
 * 	Added support for MAXine.
 * 	Fixed "bus reset" bug, which was due to the interaction
 * 	of the chained fifos and prefetching in case the target
 * 	disconnected right away on a write command.
 * 	Also reduced to a minimum the spurious interrupts at
 * 	the end of a transfer.
 * 	[92/03/02  02:17:05  af]
 * 
 * Revision 2.5  92/01/03  20:41:03  dbg
 * 	Zero dma_ptr at end of xfers, else disconnects get confused.
 * 	On disconnect, fix case where (a) it happened before we got
 * 	a chance to start the dma operations going and (b) we restart
 * 	dma on following reconnect [cuz dma_offset==1 hack].
 * 	[91/12/26  11:05:33  af]
 * 
 * Revision 2.4  91/12/13  14:55:11  jsb
 * 	Fixed to flush the FIFO after sending the command
 * 	out and before sending more data from a different place.
 * 	This should not have been necessary, but it is.
 * 	Also fixed read case to manually pick up all remaining
 * 	bytes in the SCR regs, just causing it to flush was both
 * 	incorrect and unreliable.
 * 	[91/12/12  17:49:00  af]
 * 
 * 	In kmin_dma_end_msgin(), brutally copy four bytes off the IOCTL
 * 	buffers, flushing is asynchronous and does not work well enough.
 * 	This seems to have put an end to the "bad SDP" printouts.
 * 	[91/11/01  10:53:32  af]
 * 
 * Revision 2.2.1.1  91/10/30  13:43:47  af
 * 	Added kmin_dma_end_msgin().
 * 
 * Revision 2.2  91/08/24  12:21:26  af
 * 	Actually got it to boot multiuser with on-board SCSI.
 * 	[91/08/22  11:11:55  af]
 * 
 * 	Created.
 * 	[91/08/02  03:39:00  af]
 * 
 */
/*
 *	File: kmin_dma.c
 * 	Author: Alessandro Forin, Carnegie Mellon University
 *	Date:	7/91
 *
 *	3min DMA functions, to handle the ASIC chip.
 *
 */

#include <mach/std_types.h>
#include <sys/types.h>
#include <chips/busses.h>
#include <machine/machparam.h>		/* spl definitions */
#include <device/io_req.h>
#include <device/conf.h>

#include <mips/mips_cpu.h>
#include <mips/PMAX/kmin.h>
#include <mips/PMAX/maxine.h>

#define NASIC	1	/* XXX */

#define	LANCE_DECODE		PHYS_TO_K1SEG(KMIN_REG_LANCE_DECODE)
#define	SCSI_DECODE		PHYS_TO_K1SEG(KMIN_REG_SCSI_DECODE)
#define	SCC0_DECODE		PHYS_TO_K1SEG(KMIN_REG_SCC0_DECODE)
#define	SCC1_DECODE		PHYS_TO_K1SEG(KMIN_REG_SCC1_DECODE)

#define	DTOP_DECODE		PHYS_TO_K1SEG(XINE_REG_DTOP_DECODE)
#define	FLOPPY_DECODE		PHYS_TO_K1SEG(XINE_REG_FLOPPY_DECODE)

#define	CSR			PHYS_TO_K1SEG(KMIN_REG_CSR)
#define	INTR			PHYS_TO_K1SEG(KMIN_REG_INTR)

#define	LANCE_DMAPTR		PHYS_TO_K1SEG(KMIN_REG_LANCE_DMAPTR)

#define	SCSI_DMAPTR		PHYS_TO_K1SEG(KMIN_REG_SCSI_DMAPTR)
#define	SCSI_NEXT_DMAPTR	PHYS_TO_K1SEG(KMIN_REG_SCSI_DMANPTR)
#define	SCSI_SCR		PHYS_TO_K1SEG(KMIN_REG_SCSI_SCR)

#define	DMA_ADDR(p)		(((unsigned)p) << (5-2))
#define	DMAR_READ		ASIC_CSR_SCSI_DIR	/* from SCSI to memory */

/*
 * Definition of the driver for the auto-configuration program.
 */

int	asic_probe(), asic_attach(), asic_intr();

caddr_t	asic_std[NASIC] = { 0 };
struct	bus_device *asic_info[NASIC];
struct	bus_driver asic_driver = 
        { asic_probe, 0, asic_attach, 0, asic_std, "asic", asic_info,};

/*
 * Adapt/Probe/Attach functions
 */
asic_probe( xxx, ui)
	struct bus_device *ui;
{
	/* we are only called if a 3min hand-filled us in */
	return 1;
}

asic_attach(ui)
	register struct bus_device *ui;
{
	/* we got nothing to say */
}

asic_init(isa_3min)
	boolean_t	isa_3min;
{
	volatile unsigned int	*ssr, *decoder;

	/* These are common between 3min and maxine */
	decoder = (volatile unsigned int *)LANCE_DECODE;
	*decoder = KMIN_LANCE_CONFIG;
	decoder = (volatile unsigned int *)SCSI_DECODE;
	*decoder = KMIN_SCSI_CONFIG;
	decoder = (volatile unsigned int *)SCC0_DECODE;
	*decoder = KMIN_SCC0_CONFIG;

	ssr = (volatile unsigned int *)CSR;

	if (isa_3min) {
		decoder = (volatile unsigned int *)SCC1_DECODE;
		*decoder = KMIN_SCC1_CONFIG;

		/* take all chips out of reset now */
		*ssr = 0x00000f00;

	} else {
		decoder = (volatile unsigned int *)DTOP_DECODE;
		*decoder = XINE_DTOP_CONFIG;
		decoder = (volatile unsigned int *)FLOPPY_DECODE;
		*decoder = XINE_FLOPPY_CONFIG;

		/* take all chips out of reset now */
		*ssr = 0x00001fc1;
	}

}

asic_enable_lance(phys)
	vm_offset_t	phys;
{
	volatile unsigned int	*ssr, *ldp;

	ssr = (volatile unsigned int *)CSR;
	ldp = (volatile unsigned int *)LANCE_DMAPTR;

	*ldp = (phys << 3) & 0xfff00000;

	*ssr |= ASIC_CSR_DMAEN_LANCE;
}

asic_intr(unit, spllevel, intr)
{
	if (intr & ASIC_INTR_LANCE_READ_E) {
		/* Just clear it and let the lance driver handle */
		*(volatile unsigned int *)INTR = ~ASIC_INTR_LANCE_READ_E;
	} 

	else if (intr & (ASIC_INTR_SCSI_PTR_LOAD|ASIC_INTR_SCSI_OVRUN|ASIC_INTR_SCSI_READ_E))
		kmin_dma_intr(unit, spllevel, intr);

	else
		gimmeabreak();

}


/*
 *	Functions to support DMA for the on-board SCSI
 */

#include <scsi/compat_30.h>
#include <scsi/scsi.h>
#include <scsi/scsi_defs.h>
#include <scsi/adapters/scsi_dma.h>
#include <scsi/adapters/scsi_53C94.h>

/* utilities */

#define	u_min(a,b)	(((a) < (b)) ? (a) : (b))

struct {
	vm_offset_t	kmin_dma_next_ptr;
	vm_size_t	xfer_count;
} kmin_dma_softc;

vm_offset_t	kmin_dma_gard_page;

kmin_dma_read(dma_ptr, count)
	vm_offset_t	dma_ptr;
{
	register volatile unsigned int	*ssr;
	register vm_offset_t	phys, nphys;

	ssr = (volatile unsigned int *)CSR;

	/* stop DMA engine first */
	*ssr &= ~ASIC_CSR_DMAEN_SCSI;
	* ((volatile int *)SCSI_SCR) = 0;

	phys = kvtophys(dma_ptr);
	dma_ptr += MIPS_PGBYTES;
	dma_ptr = mips_trunc_page(dma_ptr);
	nphys = kvtophys(dma_ptr);

	kmin_dma_softc.kmin_dma_next_ptr = dma_ptr;
	kmin_dma_softc.xfer_count = count;

	*(volatile int *)SCSI_DMAPTR = DMA_ADDR(phys);
	*(volatile int *)SCSI_NEXT_DMAPTR = DMA_ADDR(nphys);
	*ssr |= DMAR_READ | ASIC_CSR_DMAEN_SCSI;
}

kmin_dma_write(dma_ptr, count)
	vm_offset_t	dma_ptr;
{
	register volatile unsigned int	*ssr;
	register vm_offset_t	phys, nphys;

	ssr = (volatile unsigned int *)CSR;

	/* stop DMA engine first */
	*ssr &= ~ASIC_CSR_DMAEN_SCSI;
	* ((volatile int *)SCSI_SCR) = 0;

	phys = kvtophys(dma_ptr);
	dma_ptr += MIPS_PGBYTES;
	dma_ptr = mips_trunc_page(dma_ptr);
	nphys = kvtophys(dma_ptr);

	kmin_dma_softc.kmin_dma_next_ptr = dma_ptr;
	kmin_dma_softc.xfer_count = count;

	*(volatile int *)SCSI_DMAPTR = DMA_ADDR(phys);
	*(volatile int *)SCSI_NEXT_DMAPTR = DMA_ADDR(nphys);
	*ssr = (*ssr & ~DMAR_READ) | ASIC_CSR_DMAEN_SCSI;
}

int kmin_dma_errcnt, kmin_dma_ldcnt;

kmin_dma_intr(unit, spllevel, intr)
{
	if (intr & ASIC_INTR_SCSI_PTR_LOAD) {
		register vm_offset_t	next_phys;
		register int		togo;

		*(volatile unsigned int *)INTR = ~ASIC_INTR_SCSI_PTR_LOAD;

		/*
		 * We started with two good pointers, we just
		 * loaded one that is worth 1 page.  How much
		 * is there left to go ?
		 */
		togo = kmin_dma_softc.xfer_count - MIPS_PGBYTES;

		if (togo > MIPS_PGBYTES) {

			kmin_dma_softc.kmin_dma_next_ptr += MIPS_PGBYTES;
			next_phys = kvtophys(kmin_dma_softc.kmin_dma_next_ptr);
			*(volatile int *)SCSI_NEXT_DMAPTR =
				next_phys ? DMA_ADDR(next_phys) : -1;
		}
		/*
		 * We have all we need. Sometimes we seem to get
		 * an extra interrupt, e.g. when the target takes
		 * too long to change phase.
		 */
		else if (togo <= -MIPS_PGBYTES) {
			*((volatile unsigned int *)CSR) &= ~ASIC_CSR_DMAEN_SCSI;
		}
		/*
		 * The pointer just loaded is the last good one
		 * needed to cover the xfer.  Set sentinel.
		 */
		else {
			*(volatile int *)SCSI_NEXT_DMAPTR = kmin_dma_gard_page;
		}

		kmin_dma_softc.xfer_count = togo;
		kmin_dma_ldcnt++;
	}
	/*
	 * These interrupts are unavoidable.  There is a race between
	 * the scsi chip's interrupt and the DMA engine's one at the
	 * completion of a xfer.  Apparently, the DMA engine wins
	 * most of the times.
	 * Moreover, an invalid sentinel pointer would clash with
	 * prefetching, causing read errors.
	 */
	if (intr & (ASIC_INTR_SCSI_OVRUN|ASIC_INTR_SCSI_READ_E)) {
		*(volatile unsigned int *)INTR = ~(ASIC_INTR_SCSI_OVRUN|ASIC_INTR_SCSI_READ_E);
		kmin_dma_errcnt++;
	}
}


/*
 * Cold, probe time init
 */
opaque_t
kmin_dma_init(unit, base, dma_bsizep)
	int		unit;
	vm_offset_t	base;
	int		*dma_bsizep;
{
	kmin_dma_gard_page = DMA_ADDR(mips_trunc_page(&kmin_dma_gard_page));
	* ((volatile int *)SCSI_DMAPTR) = -1;
	* ((volatile int *)SCSI_NEXT_DMAPTR) = -1;
	* ((volatile int *)SCSI_SCR) = 0;
	return (opaque_t) 0;
}

/*
 * A target exists
 */
void
kmin_dma_new_target(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	extern vm_offset_t	vm_page_grab_phys_addr();
	vm_offset_t		buf;

	buf = vm_page_grab_phys_addr();
	if (buf == -1) panic("kmin_dma_new_target");

	/* make uncached, unless flushing turns out to be better */
	tgt->cmd_ptr = (char*)PHYS_TO_K1SEG(buf);
	tgt->dma_ptr = 0;

#ifdef	MACH_KERNEL
#else	/*MACH_KERNEL*/
	fdma_init(&tgt->fdma, scsi_per_target_virtual);
#endif	/*MACH_KERNEL*/

}

/*
 * Map virtual addresses onto physical ones
 * (and/or user virtuals on 2.5)
 */
void
kmin_dma_map(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	/*
	 * We do real DMA.
	 */
#ifdef	MACH_KERNEL
#else	/*MACH_KERNEL*/
	if (tgt->ior)
		fdma_map(&tgt->fdma, tgt->ior);
#endif	/*MACH_KERNEL*/

	if ((tgt->cur_cmd == SCSI_CMD_WRITE) ||
	    (tgt->cur_cmd == SCSI_CMD_LONG_WRITE)){
		io_req_t	ior = tgt->ior;
		register int	len = ior->io_count;

		tgt->transient_state.out_count = len;

		/* avoid leaks ?? */
		if (len < tgt->block_size) {
/*			bzero(to + len, tgt->block_size - len);*/
			len = tgt->block_size;
			tgt->transient_state.out_count = len;
		}
		/* see start_cmd & start_dataout */
		tgt->transient_state.dma_offset = 1;
	} else {
		tgt->transient_state.out_count = 0;
		tgt->transient_state.dma_offset = 0;
	}
}

int
kmin_dma_start_cmd(dma, tgt)
	opaque_t	dma;
	register target_info_t	*tgt;
{
	register int	out_count;
	vm_offset_t	phys;
	/*
	 * Note that the out_count does not include the
	 * (optional) identify message but does include
	 * both the data and cmd out count.
	 */
	out_count  = tgt->transient_state.cmd_count;
	out_count += tgt->transient_state.out_count;

	/* loadup DMA engine with DMAOUT from phys of tgt->cmd_ptr; */
	kmin_dma_write(tgt->cmd_ptr, out_count);

	return out_count;
}

void
kmin_dma_end_cmd(dma, tgt, ior)
	opaque_t	dma;
	target_info_t	*tgt;
	io_req_t	ior;
{
#ifdef	MACH_KERNEL
#else	/*MACH_KERNEL*/
	fdma_unmap(&tgt->fdma, ior);
#endif	/*MACH_KERNEL*/
	tgt->dma_ptr = 0;
	tgt->transient_state.dma_offset = 0;
}

void
kmin_dma_end_xfer(dma, tgt, bytes_read)
	opaque_t	dma;
	target_info_t	*tgt;
	int		bytes_read;
{
	register volatile char	*ptr;
	register int		nb;

	/* Was this a write op ? if so, just stop dma engine */
	if (bytes_read == 0) goto stop_dma;

	/* would be nice to do this in parallel with xfer, earlier on */
	ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;
	flush_the_damn_cache(ptr, bytes_read);

	if (nb = * ((int *)SCSI_SCR)) {
#if 0
		/* the easy way is: */
		*((int*)SCSI_SCR) = 4;
		delay(1);
		while (*((int*)SCSI_SCR)) ;
		/* but clobbers the remaining bytes (and with non-0) */
#else
		/* pick up last upto6 bytes, sigh. */
		register unsigned short	*to;
		register int w;

		/* Last byte really xferred is.. */
		ptr += bytes_read - (nb<<1);
		/* if bytes_read is odd someone is not telling the truth */
		to = (unsigned short*)(PHYS_TO_K1SEG(kvtophys(ptr)) & ~1);
		w = *(int *)(PHYS_TO_K1SEG(0x1c0401c0));
		*to++ = w;
		if (--nb > 0) {
			w >>= 16;
			*to++ = w;
		}
		if (--nb > 0) {
			w = *(int *)(PHYS_TO_K1SEG(0x1c0401d0));
			*to++ = w;
		}
#endif
	}

stop_dma:
	{
		register volatile unsigned int	*ssr;
		ssr = (volatile unsigned int *)CSR;
		*ssr &= ~ASIC_CSR_DMAEN_SCSI;
	}

	* ((volatile int *)SCSI_DMAPTR) = -1;
	* ((volatile int *)SCSI_NEXT_DMAPTR) = -1;
}

int
kmin_dma_start_datain(dma, tgt)
	opaque_t	dma;
	register target_info_t	*tgt;
{
	register char	*dma_ptr;
	register int	count;

	/* setup tgt->dma_ptr to tgt->cmd_ptr or to ior->io_data, 
	   flush cache in latter case  */
	if (tgt->cur_cmd == SCSI_CMD_READ ||
	    tgt->cur_cmd == SCSI_CMD_LONG_READ)
		dma_ptr = tgt->ior->io_data;
	else
		/* swear it won't disconnect midway a xfer or else */
		dma_ptr = tgt->cmd_ptr;
	tgt->dma_ptr = dma_ptr;

	count = tgt->transient_state.in_count;

	kmin_dma_read(dma_ptr, count);

	count = u_min(count, ASC_TC_MAX);

	return count;
}

int
kmin_dma_restart_datain_1(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	register char	*dma_ptr;
	register int	count;

	/* This is the unlikely case we disconnected before
	   actually starting the data xfer */
	if (tgt->dma_ptr == 0)
		return kmin_dma_start_datain(dma, tgt);

	/* keep dma_offset anyways */
	dma_ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;

	count = tgt->transient_state.in_count;

	kmin_dma_read(dma_ptr, count);

	count = u_min(count, ASC_TC_MAX);
	return count;
}

int
kmin_dma_restart_datain_2(dma, tgt, xferred)
	opaque_t	dma;
	target_info_t	*tgt;
	int		xferred;
{
/*XXX*/	tgt->transient_state.copy_count = tgt->transient_state.dma_offset;
	tgt->transient_state.dma_offset += xferred;

	return kmin_dma_restart_datain_1(dma, tgt);
}

void
kmin_dma_restart_datain_3(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	register volatile char	*ptr;
	int			len;

	/* avoid doing it again!! XXXX */
	ptr = tgt->dma_ptr + tgt->transient_state.copy_count;
	len = tgt->transient_state.dma_offset + tgt->transient_state.copy_count;
	flush_the_damn_cache(ptr, len);
}


boolean_t
kmin_dma_start_dataout(dma, tgt, regp, cmd)
	opaque_t	dma;
	register target_info_t	*tgt;
	volatile char	*regp, cmd;
{
	register char	*dma_ptr;

	if (tgt->transient_state.dma_offset) {

		*regp = 1;/*flush fifo*/
cmd = *regp;/*readback*/
		delay(1);

		dma_ptr = tgt->ior->io_data;
		kmin_dma_write(dma_ptr, tgt->transient_state.out_count);

	} else {
		/* do not change the current pointer, which is ok */
		dma_ptr = tgt->cmd_ptr /* + tgt->transient_state.cmd_count*/;
	}

	tgt->transient_state.dma_offset = 0;
	tgt->dma_ptr = dma_ptr;

	return TRUE;
}

int
kmin_dma_restart_dataout_1(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	int	count;

	/* this is if we disconnected before starting any xfer */
	if (tgt->dma_ptr == 0)
		kmin_dma_start_dataout(dma, tgt, &count, 0);

	count = tgt->transient_state.out_count;
	count = u_min(count, ASC_TC_MAX);

	return count;
}

int
kmin_dma_restart_dataout_2(dma, tgt, xferred)
	opaque_t	dma;
	register target_info_t	*tgt;
	int		xferred;
{
	tgt->transient_state.dma_offset += xferred;

	return kmin_dma_restart_dataout_1(dma, tgt);
}

int
kmin_dma_restart_dataout_3(dma, tgt, regp)
	opaque_t	dma;
	target_info_t	*tgt;
	volatile char	*regp;
{
	register char	*dma_ptr;
	int		ret;

	/* keep dma_offset anyways */
	dma_ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;

/*unaligned issues here, align to 8 bytes ? (maybe not..)*/
	/*
	 * See NCR spec, Pag. 12 about 'preloading the FIFO'
	 */
	if ((int)dma_ptr & 1) {
		*regp = *dma_ptr++;
		ret = 1;
	} else
		ret = 0;

	kmin_dma_write(dma_ptr, tgt->transient_state.out_count);

	return ret;
}

void
kmin_dma_restart_dataout_4(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	/* Nothing needed here */
}

int
kmin_dma_start_msgin(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	kmin_dma_read(tgt->cmd_ptr, 6);
}

void
kmin_dma_end_msgin(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	register int		*msgp;

	/*
	 * Must look at the IOCTL buffer to see if anything left
	 * Obviously yes there is stuff, else it would not be fun
	 * Note however that we expect two lonely bytes at most,
	 * so we just pick four.
	 */
	if ((* (int *)SCSI_SCR) == 0)
		return;
	msgp = (int *)tgt->cmd_ptr;
	*msgp = *(int *)(PHYS_TO_K1SEG(0x1c0401c0));
}


boolean_t
kmin_dma_disconn_1(dma, tgt, xferred)
	opaque_t	dma;
	target_info_t	*tgt;
	int		xferred;
{
	/*
	 * XXX Known bug: if we disconnect on an odd bb ??
	 */
	{
		register volatile unsigned int	*ssr;
		ssr = (volatile unsigned int *)CSR;
		*ssr &= ~ASIC_CSR_DMAEN_SCSI;
	}

	tgt->transient_state.dma_offset += xferred;
	return FALSE;
}

boolean_t
kmin_dma_disconn_2(dma, tgt)
	opaque_t	dma;
	target_info_t	*tgt;
{
	/* Nothing needed */
	return FALSE;
}


boolean_t
kmin_dma_disconn_3(dma, tgt, xferred)
	opaque_t	dma;
	target_info_t	*tgt;
	int		xferred;
{
	register volatile char	*ptr;

	{
		register volatile unsigned int	*ssr;
		ssr = (volatile unsigned int *)CSR;
		*ssr &= ~ASIC_CSR_DMAEN_SCSI;
	}

	ptr = tgt->dma_ptr + tgt->transient_state.dma_offset;
	flush_the_damn_cache(ptr, xferred);

	tgt->transient_state.dma_offset += xferred;
	return FALSE;
}

#define	kmin_dma_disconn_4		kmin_dma_disconn_3
#define	kmin_dma_disconn_5		kmin_dma_disconn_1
#define	kmin_dma_disconn_callback	0	/* never called */

flush_the_damn_cache(p,nb)
	vm_offset_t p;
{
	vm_size_t off;

	off = p & (MIPS_PGBYTES-1);
	off = MIPS_PGBYTES - off;
	while (nb) {
		off = (nb < off) ? nb : off;
		mipscache_Dflush( PHYS_TO_K0SEG(kvtophys(p)), off);
		p += off;
		nb -= off;
		off = MIPS_PGBYTES;
	}
}

scsi_dma_ops_t kmin_dma_ops = {
	kmin_dma_init,
	kmin_dma_new_target,
	kmin_dma_map,
	kmin_dma_start_cmd,
	kmin_dma_end_xfer,
	kmin_dma_end_cmd,
	kmin_dma_start_datain,
	kmin_dma_start_msgin,
	kmin_dma_end_msgin,
	kmin_dma_start_dataout,
	kmin_dma_restart_datain_1,
	kmin_dma_restart_datain_2,
	kmin_dma_restart_datain_3,
	kmin_dma_restart_dataout_1,
	kmin_dma_restart_dataout_2,
	kmin_dma_restart_dataout_3,
	kmin_dma_restart_dataout_4,
	kmin_dma_disconn_1,
	kmin_dma_disconn_2,
	kmin_dma_disconn_3,
	kmin_dma_disconn_4,
	kmin_dma_disconn_5,
	kmin_dma_disconn_callback,
};


